from typing import Self


class FileTree:
    def __init__(self, name: str, full_name: str, children: list[Self] | None = None):
        self.name = name
        self.full_name = full_name
        self.children: list[Self] = children if children is not None else []

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "full_name": self.full_name,
            "children": [child.to_dict() for child in self.children],
        }



from metripy.LangAnalyzer.Generic.Ast.AstParser import AstParser
from abc import ABC, abstractmethod


class GenericImportsAnalyzer(ABC):
    def get_import_data(
        self, filename: str, parser: AstParser
    ) -> tuple[str, list[str]]:
        return (
            self.extract_import_name(filename, parser),
            self.extract_imports(filename, parser),
        )

    @abstractmethod
    def extract_import_name(self, filename: str, parser: AstParser) -> str:
        """Extract the name of this file with which it is imported"""
        pass

    @abstractmethod
    def extract_imports(self, filename: str, parser: AstParser) -> list[str]:
        """Extract the names of the imports of this file"""
        pass



from collections import defaultdict
from datetime import datetime

from git import Repo

from metripy.Application.Config.GitConfig import GitConfig
from metripy.Metric.Git.GitMetrics import GitMetrics


class GitAnalyzer:
    def __init__(self, git_config: GitConfig):
        self.repo = Repo(git_config.repo)
        self.branch_name = git_config.branch

    def analyze(self) -> GitMetrics:
        """Main analysis method with comprehensive output"""

        # Calculate first day of this month last year
        now = datetime.now()
        first_of_month_last_year = datetime(now.year - 1, now.month, 1)
        # first_of_month_last_year = datetime(now.year, now.month, 1)
        after_date = first_of_month_last_year.strftime("%Y-%m-%d")

        return self.get_metrics(after_date)

    def _is_source_file(self, file_path: str) -> bool:
        """Check if file is a source code file we want to analyze"""
        source_extensions = {
            ".py",
            ".js",
            ".ts",
            ".tsx",
            ".php",
            ".java",
            ".cpp",
            ".c",
            ".h",
            ".rb",
            ".go",
            ".rs",
        }
        return any(file_path.endswith(ext) for ext in source_extensions)

    def get_metrics(self, after: str) -> GitMetrics:
        commits_per_month = {}
        chrun_per_month = defaultdict(lambda: {"added": 0, "removed": 0})
        file_contributors = defaultdict(lambda: {"contributors": set(), "commits": 0})
        contributor_stats = defaultdict(
            lambda: {"commits": 0, "lines_added": 0, "lines_removed": 0}
        )

        for commit in self.repo.iter_commits(
            self.branch_name, no_merges=True, after=after
        ):
            month = commit.committed_datetime.strftime("%Y-%m")
            author = commit.author.name

            if month not in commits_per_month.keys():
                commits_per_month[month] = 0
            commits_per_month[month] += 1

            stats = commit.stats.total
            insertions = stats.get("insertions", 0)
            deletions = stats.get("deletions", 0)
            chrun_per_month[month]["added"] += insertions
            chrun_per_month[month]["removed"] += deletions

            contributor_stats[author]["commits"] += 1
            contributor_stats[author]["lines_added"] += insertions
            contributor_stats[author]["lines_removed"] += deletions

            for file_path in commit.stats.files:
                if self._is_source_file(file_path):
                    file_contributors[file_path]["contributors"].add(author)
                    file_contributors[file_path]["commits"] += 1

        return GitMetrics(
            analysis_start_date=after,
            commit_stats_per_month=commits_per_month,
            churn_per_month=chrun_per_month,
            contributor_stats=contributor_stats,
            file_contributors=file_contributors,
        )



from metripy.Metric.Code.SegmentedMetrics import SegmentedMetrics
from metripy.Metric.Trend.SegmentedTrendMetric import SegmentedTrendMetric


class AggregatedTrendMetric:
    def __init__(
        self,
        historical_loc: int,
        loc: int,
        historical_avgCcPerFunction: float,
        avgCcPerFunction: float,
        historical_maintainabilityIndex: float,
        maintainabilityIndex: float,
        historical_avgLocPerFunction: float,
        avgLocPerFunction: float,
        historical_num_files: int,
        num_files: int,
        historical_segmented_loc: SegmentedMetrics,
        segmented_loc: SegmentedMetrics,
        historical_segmented_complexity: SegmentedMetrics,
        segmented_complexity: SegmentedMetrics,
        historical_segmented_maintainability: SegmentedMetrics,
        segmented_maintainability: SegmentedMetrics,
        historical_segmented_method_size: SegmentedMetrics,
        segmented_method_size: SegmentedMetrics,
        historical_avg_cog_complexity_per_function: float,
        avg_cog_complexity_per_function: float,
        historical_segmented_cog_complexity: SegmentedMetrics,
        segmented_cog_complexity: SegmentedMetrics,
    ):
        self.historical_loc = historical_loc
        self.loc_delta = loc - historical_loc
        self.historical_avgCcPerFunction = historical_avgCcPerFunction
        self.avgCcPerFunction_delta = avgCcPerFunction - historical_avgCcPerFunction
        self.historical_maintainabilityIndex = historical_maintainabilityIndex
        self.maintainabilityIndex_delta = (
            maintainabilityIndex - historical_maintainabilityIndex
        )
        self.historical_avgLocPerFunction = historical_avgLocPerFunction
        self.avgLocPerFunction_delta = avgLocPerFunction - historical_avgLocPerFunction
        self.historical_num_files = historical_num_files
        self.num_files_delta = num_files - historical_num_files

        self.historical_avg_cog_complexity_per_function = (
            historical_avg_cog_complexity_per_function
        )
        self.avg_cog_complexity_per_function_delta = (
            avg_cog_complexity_per_function - historical_avg_cog_complexity_per_function
        )

        self.historical_segmentation_data = {
            "loc": historical_segmented_loc,
            "complexity": historical_segmented_complexity,
            "cog_complexity": historical_segmented_cog_complexity,
            "maintainability": historical_segmented_maintainability,
            "methodSize": historical_segmented_method_size,
        }

        self.segmentation_data_deltas = {
            "loc": SegmentedTrendMetric(historical_segmented_loc, segmented_loc),
            "complexity": SegmentedTrendMetric(
                historical_segmented_complexity, segmented_complexity
            ),
            "cognitiveComplexity": SegmentedTrendMetric(
                historical_segmented_cog_complexity, segmented_cog_complexity
            ),
            "maintainability": SegmentedTrendMetric(
                historical_segmented_maintainability, segmented_maintainability
            ),
            "methodSize": SegmentedTrendMetric(
                historical_segmented_method_size, segmented_method_size
            ),
        }

    def get_trend_type(self, delta: float, up_is_good: bool) -> str:
        if up_is_good:
            return "positive" if delta > 0 else "negative" if delta < 0 else "neutral"
        else:
            return "negative" if delta > 0 else "positive" if delta < 0 else "neutral"

    def get_trend_icon(self, delta: float) -> str:
        return "arrow-up" if delta > 0 else "arrow-down" if delta < 0 else "arrow-right"

    def to_dict(self) -> dict:
        return {
            "loc_delta": round(self.loc_delta, 2),
            "loc_trend_type": self.get_trend_type(self.loc_delta, False),
            "loc_trend_icon": self.get_trend_icon(self.loc_delta),
            "avgCcPerFunction_delta": round(self.avgCcPerFunction_delta, 2),
            "avgCcPerFunction_trend_type": self.get_trend_type(
                self.avgCcPerFunction_delta, False
            ),
            "avgCcPerFunction_trend_icon": self.get_trend_icon(
                self.avgCcPerFunction_delta
            ),
            "avg_cog_complexity_per_function_delta": round(
                self.avg_cog_complexity_per_function_delta, 2
            ),
            "avg_cog_complexity_per_function_trend_type": self.get_trend_type(
                self.avg_cog_complexity_per_function_delta, False
            ),
            "avg_cog_complexity_per_function_trend_icon": self.get_trend_icon(
                self.avg_cog_complexity_per_function_delta
            ),
            "maintainabilityIndex_delta": round(self.maintainabilityIndex_delta, 2),
            "maintainabilityIndex_trend_type": self.get_trend_type(
                self.maintainabilityIndex_delta, True
            ),
            "maintainabilityIndex_trend_icon": self.get_trend_icon(
                self.maintainabilityIndex_delta
            ),
            "avgLocPerFunction_delta": round(self.avgLocPerFunction_delta, 2),
            "avgLocPerFunction_trend_type": self.get_trend_type(
                self.avgLocPerFunction_delta, False
            ),
            "avgLocPerFunction_trend_icon": self.get_trend_icon(
                self.avgLocPerFunction_delta
            ),
            "num_files_delta": self.num_files_delta,
            "num_files_trend_type": self.get_trend_type(self.num_files_delta, False),
            "num_files_trend_icon": self.get_trend_icon(self.num_files_delta),
        }

    def to_dict_segmentation(self) -> dict:
        return {k: v.to_dict() for k, v in self.segmentation_data_deltas.items()}



from tree_sitter import Node

from metripy.LangAnalyzer.Generic.Ast.AstParser import AstParser
from metripy.LangAnalyzer.Python.Ast.PythonAstParser import PythonAstParser


class AstDumper:
    def __init__(self, parser: AstParser):
        self.parser = parser

    def dump_all(self):
        self.dump(self.parser.tree.root_node)

    def dump(self, node: Node, depth: int = 0):
        print(" " * depth + node.type + " | " + self.parser.get_node_text(node))
        for child in node.children:
            self.dump(child, depth + 1)


if __name__ == "__main__":
    parser = PythonAstParser()
    with open("metripy/Git/GitAnalyzer.py", "r") as file:
        code = file.read()
    parser.parse(code)
    dumper = AstDumper(parser)
    dumper.dump_all()



from typing import Dict

from metripy.LangAnalyzer.Generic.CodeSmell.CodeSmell import (
    CodeSmellSeverity,
    CodeSmellType,
)
from metripy.LangAnalyzer.Generic.CodeSmell.Detectors.NamingConventionDetector import (
    NamingConventionDetector,
)


class PhpNamingConventionDetector(NamingConventionDetector):
    def _get_naming_rules(self) -> Dict[str, Dict]:
        return {
            "class": {
                "pattern": self.PATTERN_PASCAL_CASE,
                "description": "PascalCase",
                "smell_type": CodeSmellType.PASCAL_CASE_VIOLATION_CLASS,
                "severity": CodeSmellSeverity.MINOR,
                "excludes": [],
            },
            "function": {
                "pattern": self.PATTERN_CAMEL_CASE,
                "description": "camelCase",
                "smell_type": CodeSmellType.CAMEL_CASE_VIOLATION_FUNCTION,
                "severity": CodeSmellSeverity.MINOR,
                "excludes": [
                    r"^__construct$",
                    r"^__destruct$",
                    r"^__invoke$",
                ],
            },
            "variable": {
                "pattern": self.PATTERN_CAMEL_CASE,
                "description": "camelCase",
                "smell_type": CodeSmellType.CAMEL_CASE_VIOLATION_VARIABLE,
                "severity": CodeSmellSeverity.INFO,
                "excludes": [],
            },
            "constant": {
                "pattern": self.PATTERN_SCREAMING_SNAKE_CASE,
                "description": "SCREAMING_SNAKE_CASE",
                "smell_type": CodeSmellType.SCREAMING_SNAKE_CASE_VIOLATION_CONSTANT,
                "severity": CodeSmellSeverity.MINOR,
                "excludes": [],
            },
        }



from metripy.Application.Config.Config import Config
from metripy.Component.Output.CliOutput import CliOutput
from metripy.Metric.ProjectMetrics import ProjectMetrics
from metripy.Report.Cli.CodeSmellsFormatter import CodeSmellsFormatter
from metripy.Report.Cli.CouplingMetricsFormatter import CouplingMetricsFormatter
from metripy.Report.Cli.DependencyMetricsFormatter import DependencyMetricsFormatter
from metripy.Report.Cli.GitMetricsFormatter import GitMetricsFormatter
from metripy.Report.Cli.OverviewFormatter import OverviewFormatter
from metripy.Report.Cli.TopOffendersFormatter import TopOffendersFormatter
from metripy.Report.ReporterInterface import ReporterInterface


class Reporter(ReporterInterface):
    def __init__(self, config: Config, output: CliOutput):
        self.config = config
        self.output = output

    def generate(self, metrics: ProjectMetrics):
        self.output.writeln(OverviewFormatter().format(metrics))
        self.output.writeln(TopOffendersFormatter().format(metrics.file_metrics))
        self.output.writeln(CouplingMetricsFormatter().format(metrics.file_metrics))
        self.output.writeln(CodeSmellsFormatter().format(metrics.file_metrics))
        if metrics.git_metrics:
            self.output.writeln(GitMetricsFormatter().format(metrics.git_metrics))
        if metrics.dependencies:
            self.output.writeln(
                DependencyMetricsFormatter().format(metrics.dependencies)
            )



class GitConfig:
    def __init__(self):
        self.repo = "./"
        self.branch = "main"

    def to_dict(self) -> dict:
        return {
            "repo": self.repo,
            "branch": self.branch,
        }



from abc import ABC, abstractmethod
from typing import Dict


class GenericLocAnalyzer(ABC):
    """
    Abstract base class for multi-language LOC analysis.
    Subclasses must implement language-specific parsing logic.
    """

    def analyze(self, code: str) -> Dict[str, int]:
        lines = code.split("\n")
        stats = {
            "loc": len(lines),
            "sloc": 0,
            "lloc": 0,
            "comments": 0,
            "single_comments": 0,
            "multiline_comments": 0,
            "blank_lines": 0,
        }

        in_multiline_comment = False

        for line in lines:
            stripped = line.strip()

            if not stripped:
                stats["blank_lines"] += 1
                continue

            if in_multiline_comment:
                stats["multiline_comments"] += 1
                stats["comments"] += 1
                if self.is_multiline_comment_end(stripped):
                    in_multiline_comment = False
                continue

            if self.is_multiline_comment_start(stripped):
                in_multiline_comment = True
                stats["multiline_comments"] += 1
                stats["comments"] += 1
                # single line comment with mutliline markers
                if self.is_multiline_comment_end(stripped, True):
                    in_multiline_comment = False
                    stats["multiline_comments"] -= 1
                    stats["single_comments"] += 1
                continue

            if self.is_single_comment(stripped):
                stats["single_comments"] += 1
                stats["comments"] += 1
                continue

            # Count as source code
            stats["sloc"] += 1
            stats["lloc"] += self.count_logical_lines(stripped)

        return stats

    @abstractmethod
    def is_single_comment(self, line: str) -> bool:
        """Detect if a line is a single-line comment."""
        pass

    @abstractmethod
    def is_multiline_comment_start(self, line: str) -> bool:
        """Detect start of a multi-line comment."""
        pass

    @abstractmethod
    def is_multiline_comment_end(
        self, line: str, started_on_same_line: bool = False
    ) -> bool:
        """Detect end of a multi-line comment."""
        pass

    @abstractmethod
    def count_logical_lines(self, line: str) -> int:
        """Count logical statements in a line (language-specific)."""
        pass



from metripy.Tree.ClassNode import ClassNode
from metripy.Tree.FunctionNode import FunctionNode


class ModuleMetrics:
    def __init__(
        self,
        loc: int,
        avgCcPerFunction: float,
        maintainabilityIndex: float,
        avgLocPerFunction: float,
        class_nodes: list[ClassNode],
        function_nodes: list[FunctionNode],
    ):
        self.loc = loc
        self.avgCcPerFunction = avgCcPerFunction
        self.maintainabilityIndex = maintainabilityIndex
        self.avgLocPerFunction = avgLocPerFunction
        self.num_files = 1
        self.class_nodes: list[ClassNode] = class_nodes
        self.function_nodes: list[FunctionNode] = function_nodes

    def to_dict(self) -> dict:
        return {
            "loc": self.loc,
            "avgCcPerFunction": round(self.avgCcPerFunction, 2),
            "maintainabilityIndex": round(self.maintainabilityIndex, 2),
            "avgLocPerFunction": round(self.avgLocPerFunction, 2),
            "num_files": self.num_files,
            "class_nodes": [node.to_dict() for node in self.class_nodes],
            "function_nodes": [node.to_dict() for node in self.function_nodes],
        }



from metripy.Metric.Code.Segmentor import Segmentor
from metripy.Metric.ProjectMetrics import ProjectMetrics
from metripy.Report.Html.PageRenderer import PageRenderer
from metripy.Tree.FunctionNode import FunctionNode


class TopOffendersPageRenderer(PageRenderer):
    def __init__(self, template_dir: str, output_dir: str, project_name: str):
        super().__init__(template_dir, output_dir, project_name)

    def render(self, metrics: ProjectMetrics):
        orderedByTotalCc = sorted(
            metrics.file_metrics, key=lambda x: x.totalCc, reverse=True
        )[:10]
        orderedByMI = sorted(
            metrics.file_metrics, key=lambda x: x.maintainabilityIndex, reverse=False
        )[:10]
        orderedByLoc = sorted(metrics.file_metrics, key=lambda x: x.loc, reverse=True)[
            :10
        ]

        all_functions: list[FunctionNode] = []
        for fm in metrics.file_metrics:
            all_functions.extend(fm.function_nodes)

        functionsOrderedByCc = sorted(
            all_functions, key=lambda x: x.complexity, reverse=True
        )[:10]
        functionsOrderedByMi = sorted(
            all_functions, key=lambda x: x.maintainability_index, reverse=False
        )[:10]
        functionsOrderedByLoc = sorted(
            all_functions, key=lambda x: x.get_loc(), reverse=True
        )[:10]

        orderedByTotalCogCc = sorted(
            metrics.file_metrics, key=lambda x: x.total_cog_complexity, reverse=True
        )[:10]
        functionsOrderedByCogCc = sorted(
            all_functions, key=lambda x: x.cognitive_complexity, reverse=True
        )[:10]

        self.render_template(
            "top_offenders.html",
            {
                "file_loc_offenders": [
                    {**e.to_dict(), "status": Segmentor.get_loc_segment(e.loc)}
                    for e in orderedByLoc
                ],
                "file_cc_offenders": [
                    {
                        **e.to_dict(),
                        "status": Segmentor.get_complexity_segment(e.totalCc),
                    }
                    for e in orderedByTotalCc
                ],
                "file_cog_complexity_offenders": [
                    {
                        **e.to_dict(),
                        "status": Segmentor.get_complexity_segment(
                            e.total_cog_complexity
                        ),
                    }
                    for e in orderedByTotalCogCc
                ],
                "file_mi_offenders": [
                    {
                        **e.to_dict(),
                        "status": Segmentor.get_maintainability_segment(
                            e.maintainabilityIndex
                        ),
                    }
                    for e in orderedByMI
                ],
                "function_size_offenders": [
                    {
                        **e.to_dict(),
                        "status": Segmentor.get_method_size_segment(e.get_loc()),
                    }
                    for e in functionsOrderedByLoc
                ],
                "function_cc_offenders": [
                    {
                        **e.to_dict(),
                        "status": Segmentor.get_complexity_segment(e.complexity),
                    }
                    for e in functionsOrderedByCc
                ],
                "function_cog_complexity_offenders": [
                    {
                        **e.to_dict(),
                        "status": Segmentor.get_complexity_segment(
                            e.cognitive_complexity
                        ),
                    }
                    for e in functionsOrderedByCogCc
                ],
                "function_mi_offenders": [
                    {
                        **e.to_dict(),
                        "status": Segmentor.get_maintainability_segment(
                            e.maintainability_index
                        ),
                    }
                    for e in functionsOrderedByMi
                ],
            },
        )



from metripy.LangAnalyzer.Generic.Metrics.GenericCyclomaticComplexityAnalyzer import (
    GenericCyclomaticComplexityAnalyzer,
)
from metripy.LangAnalyzer.Php.Metrics.PhpCyclomaticComplexityAnalyzer import (
    PhpCyclomaticComplexityAnalyzer,
)
from metripy.LangAnalyzer.Python.Metrics.PythonCyclomaticComplexityAnalyzer import (
    PythonCyclomaticComplexityAnalyzer,
)


class CyclomaticComplexityAnalyzerFactory:
    _ANALYZERS = {
        "Python": PythonCyclomaticComplexityAnalyzer(),
        "PHP": PhpCyclomaticComplexityAnalyzer(),
        "Typescript": None,
    }

    @staticmethod
    def get_cyclomatic_complexity_analyzer(
        language: str,
    ) -> GenericCyclomaticComplexityAnalyzer:
        try:
            return CyclomaticComplexityAnalyzerFactory._ANALYZERS[language]
        except KeyError:
            raise ValueError(
                f"No cyclomatic complexity analyzer found for language: {language}"
            )



from metripy.LangAnalyzer.Generic.CodeSmell.CodeSmell import CodeSmell
from metripy.Tree.ClassNode import ClassNode
from metripy.Tree.FunctionNode import FunctionNode


class ModuleNode:
    def __init__(
        self,
        full_name: str,
        loc: int,
        lloc: int,
        sloc: int,
        comments: int,
        multi: int,
        blank: int,
        single_comments: int,
    ):
        self.full_name = full_name
        self.loc = loc
        self.lloc = lloc
        self.sloc = sloc
        self.comments = comments
        self.multi = multi
        self.blank = blank
        self.single_comments = single_comments
        self.maintainability_index = 0
        self.classes: list[ClassNode] = []
        self.functions: list[FunctionNode] = []
        self.imports: list[str] | None = None
        self.import_name: str | None = None
        self.code_smells: list[CodeSmell] = []

    def to_dict(self) -> dict:
        return {
            "full_name": self.full_name,
            "loc": self.loc,
            "lloc": self.lloc,
            "sloc": self.sloc,
            "comments": self.comments,
            "multi": self.multi,
            "blank": self.blank,
            "single_comments": self.single_comments,
            "maintainability_index": self.maintainability_index,
            "classes": [c.to_dict() for c in self.classes],
            "functions": [f.to_dict() for f in self.functions],
            "imports": self.imports,
            "import_name": self.import_name,
            "code_smells": [c.to_dict() for c in self.code_smells],
        }



"""Generic code smell detector that works with any language parser"""

from typing import List, Type

from metripy.Application.Config.CodeSmellConfig import CodeSmellConfig
from metripy.LangAnalyzer.Generic.Ast.AstParser import AstParser
from metripy.LangAnalyzer.Generic.CodeSmell.CodeSmell import CodeSmell
from metripy.LangAnalyzer.Generic.CodeSmell.Detectors.BaseDetector import (
    BaseCodeSmellDetector,
)
from metripy.LangAnalyzer.Generic.CodeSmell.Detectors.TooManyParametersDetector import (
    TooManyParametersDetector,
)
from metripy.LangAnalyzer.Generic.CodeSmell.Detectors.UnusedFunctionsDetector import (
    UnusedFunctionsDetector,
)
from metripy.LangAnalyzer.Generic.CodeSmell.Detectors.UnusedImportsDetector import (
    UnusedImportsDetector,
)


class GenericCodeSmellDetector:
    """
    Generic code smell detector that orchestrates multiple detectors.
    Works with any language that has an AstParser implementation.
    """

    def __init__(
        self,
        config: CodeSmellConfig,
        filename: str,
        code: str,
        parser: AstParser,
        naming_detector_class: Type[BaseCodeSmellDetector] | None = None,
        max_params: int = 5,
    ):
        """
        Initialize generic code smell detector.

        Args:
            filename: Path to the file being analyzed
            code: Source code to analyze
            parser: Language-specific AST parser
            naming_detector_class: Optional language-specific naming convention detector
            max_params: Maximum number of parameters for functions
        """
        self.config = config
        self.filename = filename
        self.code = code
        self.parser = parser
        self.naming_detector_class = naming_detector_class
        self.max_params = max_params

        # Parse the code
        self.parser.parse(code)

    def detect_all(self) -> List[CodeSmell]:
        """Run all code smell detection checks"""
        if self.parser.tree is None:
            return []

        all_smells = []

        # Create and run generic detectors
        detectors: dict[str, BaseCodeSmellDetector] = {
            "unused_imports": UnusedImportsDetector(
                self.filename, self.parser, self.config
            ),
            "unused_functions": UnusedFunctionsDetector(
                self.filename, self.parser, self.config
            ),
            "too_many_parameters": TooManyParametersDetector(
                self.filename, self.parser, self.config, self.max_params
            ),
        }

        # Add naming convention detector if provided
        if self.naming_detector_class:
            detectors["naming_convention"] = self.naming_detector_class(
                self.filename, self.parser, self.config
            )

        # Run all configured detectors
        for detector_name, detector in detectors.items():
            if self.config.is_enabled(detector_name):
                all_smells.extend(detector.detect())

        return all_smells

    @staticmethod
    def for_python(
        config: CodeSmellConfig, filename: str, code: str
    ) -> "GenericCodeSmellDetector":
        """Factory method to create detector for Python"""
        from metripy.LangAnalyzer.Python.Ast.PythonAstParser import PythonAstParser
        from metripy.LangAnalyzer.Python.CodeSmell.PythonNamingConventionDetector import (
            PythonNamingConventionDetector,
        )

        parser = PythonAstParser()
        return GenericCodeSmellDetector(
            config,
            filename,
            code,
            parser,
            naming_detector_class=PythonNamingConventionDetector,
        )

    @staticmethod
    def for_php(
        config: CodeSmellConfig, filename: str, code: str
    ) -> "GenericCodeSmellDetector":
        """Factory method to create detector for PHP"""
        from metripy.LangAnalyzer.Php.Ast.PhpAstParser import PhpAstParser
        from metripy.LangAnalyzer.Php.CodeSmell.PhpNamingConventionDetector import (
            PhpNamingConventionDetector,
        )

        parser = PhpAstParser()
        return GenericCodeSmellDetector(
            config,
            filename,
            code,
            parser,
            naming_detector_class=PhpNamingConventionDetector,
        )

    @staticmethod
    def for_typescript(
        config: CodeSmellConfig, filename: str, code: str
    ) -> "GenericCodeSmellDetector":
        """Factory method to create detector for TypeScript"""
        from metripy.LangAnalyzer.Typescript.Ast.TypescriptAstParser import (
            TypescriptAstParser,
        )
        from metripy.LangAnalyzer.Typescript.CodeSmell.TypescriptNamingConventionDetector import (
            TypescriptNamingConventionDetector,
        )

        parser = TypescriptAstParser()
        return GenericCodeSmellDetector(
            config,
            filename,
            code,
            parser,
            naming_detector_class=TypescriptNamingConventionDetector,
        )



from metripy.Metric.ProjectMetrics import ProjectMetrics
from metripy.Report.Html.PageRenderer import PageRenderer


class MetricsPageRenderer(PageRenderer):
    def __init__(self, template_dir: str, output_dir: str, project_name: str):
        super().__init__(template_dir, output_dir, project_name)

    def render(self, metrics: ProjectMetrics):
        # only showing static info
        self.render_template("metrics.html", {})



from typing import List

from metripy.Application.Config.CodeSmellConfig import CodeSmellConfig
from metripy.LangAnalyzer.Generic.CodeSmell.CodeSmell import CodeSmell
from metripy.LangAnalyzer.Generic.CodeSmell.GenericCodeSmellDetector import (
    GenericCodeSmellDetector,
)


class PhpCodeSmellDetector:
    def __init__(self, config: CodeSmellConfig):
        self.config = config

    def detect_all(self, filename: str, code: str) -> List[CodeSmell]:
        detector = GenericCodeSmellDetector.for_php(self.config, filename, code)
        return detector.detect_all()



from metripy.LangAnalyzer.Generic.Metrics.GenericLocAnalyzer import GenericLocAnalyzer


class PythonLocAnalyzer(GenericLocAnalyzer):
    def is_single_comment(self, line: str) -> bool:
        return line.startswith("#")

    def is_multiline_comment_start(self, line: str) -> bool:
        return line.startswith('"""') or line.startswith("'''")

    def is_multiline_comment_end(
        self, line: str, started_on_same_line: bool = False
    ) -> bool:
        if started_on_same_line:
            return (line.endswith('"""') or line.endswith("'''")) and len(line) > 3
        return line.endswith('"""') or line.endswith("'''")

    def count_logical_lines(self, line: str) -> int:
        # Rough heuristic: split by ';'
        return len([stmt for stmt in line.split(";") if stmt.strip()])



import re


class PhpBasicAstParser:

    @staticmethod
    def parse_php_structure(code: str):

        # Regex patterns
        class_pattern = re.compile(r"class\s+(\w*)")
        interface_pattern = re.compile(r"interface\s+(\w*)")
        method_pattern = re.compile(r"function\s+(\w+)\s*\([^)]*\)?")
        function_pattern = re.compile(r"function\s+(\w+)\s*\([^)]*\)?")

        lines = code.split("\n")
        structure = []
        current_class = None

        for i, line in enumerate(lines):
            class_match = class_pattern.search(line)
            if class_match:
                current_class = class_match.group(1)
                structure.append(
                    {"type": "class", "name": current_class, "line": i + 1}
                )

            interface_match = interface_pattern.search(line)
            if interface_match:
                current_class = interface_match.group(1)
                structure.append(
                    {
                        "type": "class",
                        "type_type": "interface",
                        "name": current_class,
                        "line": i + 1,
                    }
                )

            method_match = method_pattern.search(line)
            if method_match and current_class:
                structure.append(
                    {
                        "type": "method",
                        "name": method_match.group(1),
                        "class": current_class,
                        "line": i + 1,
                    }
                )

            elif function_pattern.search(line) and not current_class:
                function_name = function_pattern.search(line).group(1)
                structure.append(
                    {"type": "function", "name": function_name, "line": i + 1}
                )

        return structure



from metripy.Application.Config.ProjectConfig import ProjectConfig
from metripy.Component.Debug.Debugger import Debugger
from metripy.Component.Output.CliOutput import CliOutput
from metripy.Component.Output.ProgressBar import ProgressBar
from metripy.Dependency.Composer.Composer import Composer
from metripy.Dependency.Dependency import Dependency
from metripy.Dependency.Npm.Npm import Npm
from metripy.Dependency.Pip.Pip import Pip
from metripy.Git.GitAnalyzer import GitAnalyzer
from metripy.Import.Json.JsonImporter import JsonImporter
from metripy.LangAnalyzer.AbstractLangAnalyzer import AbstractLangAnalyzer
from metripy.LangAnalyzer.Php.PhpAnalyzer import PhpAnalyzer
from metripy.LangAnalyzer.Python.PythonAnalyzer import PythonAnalyzer
from metripy.LangAnalyzer.Typescript.TypescriptAnalyzer import TypescriptAnalyzer
from metripy.Metric.Code.FileMetrics import FileMetrics
from metripy.Metric.Git.GitMetrics import GitMetrics
from metripy.Metric.ProjectMetrics import ProjectMetrics
from metripy.Trend.TrendAnalyzer import TrendAnalyzer


class Analyzer:
    def __init__(self, config: ProjectConfig, output: CliOutput, debugger: Debugger):
        self.config = config
        self.output = output
        self.debugger = debugger
        self.runners: list[AbstractLangAnalyzer] = [
            PythonAnalyzer(self.config),
            PhpAnalyzer(self.config),
            TypescriptAnalyzer(self.config),
        ]

    def analyze_git(self) -> GitMetrics:
        self.output.writeln("<info>Analyzing git history...</info>")
        metrics = GitAnalyzer(self.config.git).analyze()
        self.output.writeln("<success>Git history analyzed</success>")

        return metrics

    def analyze_code(self, files: list[str]) -> list[FileMetrics]:
        file_metrics = []

        # for multi language projects, register runner per language
        runner_sizes = []
        for runner in self.runners:
            runner.set_files(files)
            if not runner.is_needed():
                continue
            runner_sizes.append(len(runner.files))

        self.debugger.debug(len(files))
        self.output.writeln("<info>Analyzing code...</info>")
        progress_bar = ProgressBar(self.output, sum(runner_sizes))
        for runner in self.runners:
            if not runner.is_needed():
                continue
            runner.before_run()
            runner.run(progress_bar)
            runner.after_run()

            for metric in runner.get_metrics():
                file_metrics.append(metric)

        self.output.writeln("")
        self.output.writeln("<success>Code analyzed</success>")

        return file_metrics

    def analyze_composer(self) -> list[Dependency]:
        self.output.writeln("<info>Analyzing composer packages...</info>")
        dependencies = Composer().get_composer_dependencies(self.config.base_path)
        self.output.writeln("<success>Composer packages analyzed</success>")

        return dependencies

    def analyze_pip(self) -> list[Dependency]:
        self.output.writeln("<info>Analyzing pip packages...</info>")
        dependencies = Pip().get_dependencies(self.config.base_path)
        self.output.writeln("<success>Pip packages analyzed</success>")

        return dependencies

    def analyze_npm(self) -> list[Dependency]:
        self.output.writeln("<info>Analyzing npm packages...</info>")
        dependencies = Npm().get_dependencies(self.config.base_path)
        self.output.writeln("<success>Npm packages analyzed</success>")

        return dependencies

    def add_trends(self, project_metrics: ProjectMetrics):
        self.output.writeln("<info>Analyzing trends...</info>")
        importer = JsonImporter(self.output)
        historical_project_metrics = importer.import_data(self.config.history_path)
        TrendAnalyzer().add_historical_file_trends(
            project_metrics.file_metrics, historical_project_metrics.file_metrics
        )
        TrendAnalyzer().add_historical_project_trends(
            project_metrics, historical_project_metrics
        )
        self.output.writeln("<success>Trends analyzed</success>")

    def run(self, files: list[str]) -> ProjectMetrics:
        git_stats = None
        if self.config.git:
            git_stats = self.analyze_git()

        file_metrics = []
        if True:
            file_metrics = self.analyze_code(files)

        # TODO: analyze dependencies: composer, pip, npm, etc etc
        packages = None
        if self.config.composer:
            packages = self.analyze_composer()
        elif self.config.pip:
            packages = self.analyze_pip()
        elif self.config.npm:
            packages = self.analyze_npm()

        if not self.config.history_path:
            return ProjectMetrics(file_metrics, git_stats, packages)

        # analyze trends
        project_metrics = ProjectMetrics(file_metrics, git_stats, packages)
        self.add_trends(project_metrics)

        return project_metrics



import re
import sys
from typing import Self


class CliOutput:
    def __init__(self):
        self.quiet_mode = False

    def set_quiet_mode(self, quiet_mode: bool):
        self.quiet_mode = quiet_mode

    def writeln(self, message: str) -> Self:
        self.write(str(message), end="\n")

        return self

    def write(self, message: str, end="") -> Self:
        if matches := re.search(r"<([a-z]+)>(.*?)</([a-z]+)>", message):
            type = matches.group(1)
            message = matches.group(2)
            color = {
                "error": "\033[31m",
                "warning": "\033[33m",
                "success": "\033[32m",
                "info": "\033[34m",
                "debug": "\033[95m",
            }
            message = color[type] + message + "\033[0m"

        if not self.quiet_mode:
            print(message, end=end)

        return self

    def err(self, message: str) -> Self:
        sys.stderr.write(message)

        return self

    def clearln(self) -> Self:
        if self.has_ansi():
            self.write("\x0d")
            self.write("\x1b[2K")

        return self

    def has_ansi(self) -> bool:
        return sys.stdout.isatty()



from metripy.LangAnalyzer.Generic.Metrics.GenericImportsAnalyzer import (
    GenericImportsAnalyzer,
)
from metripy.LangAnalyzer.Python.PythonImportsAnalyzer import PythonImportsAnalyzer
from metripy.LangAnalyzer.Php.PhpImportsAnalyzer import PhpImportsAnalyzer


class ImportsAnalyzerFactory:
    _ANALYZERS = {
        "Python": PythonImportsAnalyzer(),
        "PHP": PhpImportsAnalyzer(),
        "Typescript": None,
    }

    @staticmethod
    def get_imports_analyzer(language: str) -> GenericImportsAnalyzer:
        try:
            return ImportsAnalyzerFactory._ANALYZERS[language]
        except KeyError:
            raise ValueError(f"No imports analyzer found for language: {language}")



import os

from metripy.Application.Config.ProjectConfig import ProjectConfig


class Finder:
    def fetch(self, project_configs: list[ProjectConfig]) -> dict[str, list[str]]:
        """returns a list of files per project project_name => [files,...]"""
        project_files = {}
        for project_config in project_configs:
            files = []

            paths = [
                os.path.join(project_config.base_path, inc)
                for inc in project_config.includes
            ]
            for path in paths:
                if os.path.isdir(path):
                    self._search(
                        path, project_config.extensions, project_config.excludes, files
                    )
                elif path not in project_config.excludes and path.endswith(
                    tuple(project_config.extensions)
                ):
                    files.append(path)
            project_files[project_config.name] = files

        return project_files

    def _search(
        self, path: str, extensions: list[str], excludes: list[str], results: list[str]
    ):
        if not os.path.isdir(path):
            if path not in excludes and path.endswith(tuple(extensions)):
                results.append(path)

        for file in os.listdir(path):
            if os.path.isdir(os.path.join(path, file)) and file not in excludes:
                self._search(os.path.join(path, file), extensions, excludes, results)
            elif file not in excludes and file.endswith(tuple(extensions)):
                results.append(os.path.join(path, file))



import json

from metripy.Metric.ProjectMetrics import ProjectMetrics
from metripy.Report.Html.PageRenderer import PageRenderer


class GitAnalysisPageRenderer(PageRenderer):
    def __init__(self, template_dir: str, output_dir: str, project_name: str):
        super().__init__(template_dir, output_dir, project_name)

    def _render_empty_template(self):
        self.render_template(
            "git_analysis.html",
            {
                "has_git_analysis_data": False,
                "git_analysis": {},
                "git_analysis_json": "{}",
                "git_stats_data": "{}",
                "git_churn_data": "{}",
                "git_silos_data": [],
                "git_contributors": [],
                "git_hotspots_data": [],
            },
        )

    def render(self, metrics: ProjectMetrics):
        if not metrics.git_metrics:
            self._render_empty_template()
            return

        self.render_template(
            "git_analysis.html",
            {
                "has_git_analysis_data": bool(metrics.git_metrics),
                "git_analysis": metrics.git_metrics.to_dict(),
                "git_analysis_json": json.dumps(
                    metrics.git_metrics.get_contributors_dict(), indent=4
                ),
                "git_stats_data": json.dumps(
                    metrics.git_metrics.get_commit_stats_per_month(), indent=4
                ),  # git commit graph
                "git_churn_data": json.dumps(
                    metrics.git_metrics.get_churn_per_month(), indent=4
                ),  # git chrun graph
                "git_silos_data": metrics.git_metrics.get_silos_list()[
                    :10
                ],  # silos list
                "git_contributors": metrics.git_metrics.get_contributors_list()[
                    :10
                ],  # contributors list
                "git_hotspots_data": metrics.git_metrics.get_hotspots_list()[
                    :10
                ],  # hotspots list
            },
        )



class ClassTrendMetric:
    def __init__(
        self,
        historical_lineno: int,
        lineno: int,
        historical_real_complexity: int,
        real_complexity: int,
    ):
        self.historical_lineno = historical_lineno
        self.lineno_delta = lineno - historical_lineno
        self.historical_real_complexity = historical_real_complexity
        self.real_complexity_delta = real_complexity - historical_real_complexity

    def to_dict(self) -> dict:
        return {
            "historical_lineno": self.historical_lineno,
            "lineno_delta": self.lineno_delta,
            "historical_real_complexity": self.historical_real_complexity,
            "real_complexity_delta": self.real_complexity_delta,
        }



import math
from typing import Dict

from tree_sitter import Node

from metripy.LangAnalyzer.Generic.Ast.AstParser import AstParser


class HalsteadAnalyzer:
    def __init__(self, parser: AstParser):
        self.parser = parser
        self.operators = []
        self.operands = []

    def analyze(self) -> Dict[str, float]:
        self._collect_tokens(self.parser.tree.root_node)

        n1 = len(set(self.operators))
        n2 = len(set(self.operands))
        N1 = len(self.operators)
        N2 = len(self.operands)

        n = n1 + n2
        N = N1 + N2
        V = N * (0 if n == 0 else (math.log2(n)))
        D = (n1 / 2) * (N2 / n2 if n2 else 0)
        E = D * V

        return {
            "n1": n1,
            "n2": n2,
            "N1": N1,
            "N2": N2,
            "Vocabulary": n,
            "Length": N,
            "Volume": V,
            "Difficulty": D,
            "Effort": E,
        }

    def _collect_tokens(self, node: Node):
        # Language-specific classification
        if node.type in self.parser.get_operator_types():
            self.operators.append(self.parser.get_node_text(node))
        elif node.type in self.parser.get_operand_types():
            self.operands.append(self.parser.get_node_text(node))

        for child in node.children:
            self._collect_tokens(child)



from metripy.Application.Config.Config import Config
from metripy.Component.Output.CliOutput import CliOutput
from metripy.Metric.ProjectMetrics import ProjectMetrics
from metripy.Report.Json.AbstractJsonReporter import AbstractJsonReporter


class JsonReporter(AbstractJsonReporter):
    def __init__(self, config: Config, output: CliOutput):
        self.config = config
        self.output = output

    def generate(self, metrics: ProjectMetrics):
        self.put_data(metrics.to_dict())
        self.output.writeln(
            f"<success>Create json report in {self.config.path}</success>"
        )



from typing import Self

from metripy.Component.Output.CliOutput import CliOutput


class Debugger:
    def __init__(self, output: CliOutput):
        self.enabled = False
        self.output = output

    def enable(self) -> Self:
        self.enabled = True

        return self

    def debug(self, message: str):
        if not self.enabled:
            return

        self.output.writeln(f"<debug>{message}</debug>")



"""
Python code smell detector - wrapper around the generic tree-sitter-based detector.

For backward compatibility, this provides the same interface as before.
Now uses the generic GenericCodeSmellDetector framework.
"""

from typing import List

from metripy.Application.Config.CodeSmellConfig import CodeSmellConfig
from metripy.LangAnalyzer.Generic.CodeSmell.CodeSmell import CodeSmell
from metripy.LangAnalyzer.Generic.CodeSmell.GenericCodeSmellDetector import (
    GenericCodeSmellDetector,
)


class PythonCodeSmellDetector:
    """
    Python code smell detector using tree-sitter-based generic framework.

    This is a convenience wrapper that maintains backward compatibility.
    For direct access, use: GenericCodeSmellDetector.for_python()
    """

    def __init__(self, config: CodeSmellConfig):
        self.config = config

    def detect_all(self, filename: str, code: str) -> List[CodeSmell]:
        """Run all configured code smell detection checks"""
        detector = GenericCodeSmellDetector.for_python(self.config, filename, code)
        return detector.detect_all()



import os

from metripy.Application.Config.Config import Config
from metripy.Component.Output.CliOutput import CliOutput
from metripy.Metric.ProjectMetrics import ProjectMetrics
from metripy.Tree.FunctionNode import FunctionNode


class Reporter:

    HEADERS = [
        "File",
        "Class",
        "Function",
        "LOC",
        "Cyclomatic Complexity",
        "Cognitive Complexity",
        "Maintainability Index",
        "Halstead h1",
        "Halstead h2",
        "Halstead n1",
        "Halstead n2",
        "Vocabulary",
        "Length",
        "Calculated Length",
        "Volume",
        "Difficulty",
        "Effort",
        "Time",
        "Bugs",
    ]

    def __init__(self, config: Config, output: CliOutput):
        self.config = config
        self.output = output

        self.delimiter = ","

        os.makedirs(os.path.dirname(self.config.path), exist_ok=True)
        self.file = open(self.config.path, "w")

    def __del__(self):
        self.file.close()

    def _write_line(self, data: list[str]):
        self.file.write(self.delimiter.join(data) + "\n")

    def generate(self, metrics: ProjectMetrics):
        self._write_line(self.HEADERS)

        data = []
        already_seen_functions = set()
        for file in metrics.file_metrics:
            for class_node in file.class_nodes:
                for function in class_node.functions:
                    data.append(
                        self.create_function_row(
                            file.full_name, class_node.full_name, function
                        )
                    )
                    already_seen_functions.add(function.full_name)
            # to catch global functions
            for function in file.function_nodes:
                if function.full_name in already_seen_functions:
                    continue
                data.append(self.create_function_row(file.full_name, "", function))
                already_seen_functions.add(function.full_name)

        for row in data:
            self._write_line(row)

    def create_function_row(
        self, file_name: str, class_name: str, function: FunctionNode
    ) -> list[str]:
        return [
            str(elem)
            for elem in [
                file_name,
                class_name,
                function.full_name,
                function.get_loc(),
                function.complexity,
                function.cognitive_complexity,
                function.maintainability_index,
                function.h1,
                function.h2,
                function.N1,
                function.N2,
                function.vocabulary,
                function.length,
                function.calculated_length,
                function.volume,
                function.difficulty,
                function.effort,
                function.time,
                function.bugs,
            ]
        ]



from metripy.Metric.FileTree.FileTree import FileTree


class FileTreeParser:
    @staticmethod
    def parse(paths: list[str], shorten: bool = False) -> FileTree:
        root = FileTree(".", ".")

        for path in paths:
            parts = path.strip("./").split("/")
            current = root

            for part in parts:
                # Check if part already exists in current children
                found = next(
                    (child for child in current.children if child.name == part), None
                )
                if not found:
                    found = FileTree(part, path)
                    current.children.append(found)
                current = found

        if shorten:
            FileTreeParser._shorten_tree(root)

        return root

    @staticmethod
    def _shorten_tree(node: FileTree):
        """shorten tree nodes that only have a single child"""
        while len(node.children) == 1:
            child = node.children[0]
            node.name += "/" + child.name
            node.full_name = child.full_name
            node.children = child.children

        for child in node.children:
            FileTreeParser._shorten_tree(child)



"""Generic detector for functions with too many parameters"""

from typing import List

from metripy.Application.Config.CodeSmellConfig import CodeSmellConfig
from metripy.LangAnalyzer.Generic.CodeSmell.CodeSmell import (
    CodeSmell,
    CodeSmellSeverity,
    CodeSmellType,
)
from metripy.LangAnalyzer.Generic.CodeSmell.Detectors.BaseDetector import (
    BaseCodeSmellDetector,
)


class TooManyParametersDetector(BaseCodeSmellDetector):
    """Detects functions with too many parameters in any language"""

    def __init__(
        self, filename: str, parser, config: CodeSmellConfig, max_params: int = 5
    ):
        super().__init__(filename, parser, config)
        self.max_params = max_params
        # Language-specific parameters to exclude
        self.excluded_params = {"self", "cls", "this"}

    def detect(self) -> List[CodeSmell]:
        """Detect functions with too many parameters"""
        function_nodes = self.parser.get_function_nodes()
        if not function_nodes:
            return []

        for func_node in function_nodes:
            func_name = self.parser.extract_function_name(func_node)
            if not func_name:
                continue

            params = self.parser.get_function_parameters(func_node)
            # Filter out language-specific parameters like 'self', 'this'
            filtered_params = [p for p in params if p not in self.excluded_params]
            param_count = len(filtered_params)

            if param_count > self.max_params:

                self.smells.append(
                    CodeSmell(
                        smell_type=CodeSmellType.TOO_MANY_PARAMETERS,
                        severity=CodeSmellSeverity.MAJOR,
                        filename=self.filename,
                        line_number=self.get_line_number(func_node),
                        column=self.get_column(func_node),
                        code_line=self.parser.get_code_line(func_node),
                        message=f"Function '{func_name}' has {param_count} parameters (max recommended: {self.max_params})",
                        symbol=func_name,
                    )
                )

        return self.smells



"""TypeScript-specific AST parser implementation using tree-sitter"""

from typing import List, Optional

from tree_sitter import Node

from metripy.LangAnalyzer.Generic.Ast.AstParser import AstParser


class TypescriptAstParser(AstParser):
    """TypeScript-specific implementation of AST parser"""

    def __init__(self):
        super().__init__("typescript")

    def get_import_nodes(self) -> List[Node]:
        """Get all import statements"""
        return self.find_nodes_by_type("import_statement")

    def get_function_nodes(self) -> List[Node]:
        """Get all function and method definitions"""
        functions = self.find_nodes_by_type("function_declaration")
        functions.extend(self.find_nodes_by_type("method_definition"))
        functions.extend(self.find_nodes_by_type("arrow_function"))
        return functions

    def get_class_nodes(self) -> List[Node]:
        """Get all class definitions"""
        return self.find_nodes_by_type("class_declaration")

    def get_variable_assignment_nodes(self) -> List[Node]:
        """Get all variable declarations"""
        nodes = self.find_nodes_by_type("lexical_declaration")
        nodes.extend(self.find_nodes_by_type("variable_declaration"))
        return nodes

    def get_identifier_nodes(self, context: str = "usage") -> List[Node]:
        """Get identifier nodes"""
        return self.find_nodes_by_type("identifier")

    def extract_function_name(self, node: Node) -> Optional[str]:
        """Extract function name"""
        if node.type == "function_declaration":
            for child in node.children:
                if child.type == "identifier":
                    return self.get_node_text(child)
        elif node.type == "method_definition":
            for child in node.children:
                if child.type == "property_identifier":
                    return self.get_node_text(child)
        elif node.type == "arrow_function":
            # Arrow functions might not have a direct name
            # Need to look at parent context
            return None
        return None

    def extract_class_name(self, node: Node) -> Optional[str]:
        """Extract class name"""
        for child in node.children:
            if child.type == "type_identifier":
                return self.get_node_text(child)
        return None

    def extract_variable_name(self, node: Node) -> Optional[str]:
        """Extract variable name from declaration"""
        for child in node.children:
            if child.type == "variable_declarator":
                for grandchild in child.children:
                    if grandchild.type == "identifier":
                        return self.get_node_text(grandchild)
        return None

    def extract_import_name(self, node: Node) -> Optional[str]:
        """Extract imported names from import statement"""
        names = []
        for child in node.children:
            if child.type == "import_clause":
                for grandchild in child.children:
                    if grandchild.type == "identifier":
                        names.append(self.get_node_text(grandchild))
                    elif grandchild.type == "named_imports":
                        for import_spec in grandchild.children:
                            if import_spec.type == "import_specifier":
                                for part in import_spec.children:
                                    if part.type == "identifier":
                                        names.append(self.get_node_text(part))
        return ", ".join(names) if names else None

    def get_function_parameters(self, function_node: Node) -> List[str]:
        """Extract parameter names from function definition"""
        params = []
        for child in function_node.children:
            if child.type == "formal_parameters":
                for param_node in child.children:
                    if (
                        param_node.type == "required_parameter"
                        or param_node.type == "optional_parameter"
                    ):
                        for grandchild in param_node.children:
                            if grandchild.type == "identifier":
                                params.append(self.get_node_text(grandchild))
                                break
                    elif param_node.type == "identifier":
                        params.append(self.get_node_text(param_node))
        return params

    def get_operator_types(self) -> List[str]:
        raise NotImplementedError

    def get_operand_types(self) -> List[str]:
        raise NotImplementedError

    def get_function_attributes(self, function_node: Node) -> List[str]:
        raise NotImplementedError

    def get_function_self_calls(self, function_node: Node) -> List[str]:
        raise NotImplementedError

    def get_class_methods(self, class_node: Node) -> List[Node]:
        raise NotImplementedError

    def extract_import_qualified_name(self, node: Node) -> Optional[str]:
        raise NotImplementedError



from abc import ABC, abstractmethod

from metripy.Application.Config.Config import Config
from metripy.Application.Config.File.PathResolver import PathResolver
from metripy.Application.Config.GitConfig import GitConfig
from metripy.Application.Config.ProjectConfig import ProjectConfig
from metripy.Application.Config.ReportConfig import ReportConfig


class ConfigFileReaderInterface(ABC):
    def __init__(self, filename: str):
        self.filename = filename

    @abstractmethod
    def read(self, config: Config) -> None:
        pass

    def resolve_path(self, path: str) -> str:
        return PathResolver.resolve(path)

    def parse_data(self, data: dict, config: Config) -> None:
        # configs
        if configs := data.get("configs"):
            for project_name, project_config in configs.items():
                project_config = self.parse_config_json(project_name, project_config)
                config.project_configs.append(project_config)

    def parse_config_json(self, project_name: str, data: dict) -> ProjectConfig:
        project_config = ProjectConfig(project_name)

        # extensions
        if base_path := data.get("base_path"):
            project_config.base_path = base_path

        # includes
        if includes := data.get("includes"):
            files = []
            # with config file, includes are relative to the config file
            for include in includes:
                include = self.resolve_path(include)
                files.append(include)

            project_config.includes = files

        # extensions
        if extensions := data.get("extensions"):
            project_config.extensions = extensions

        # excludes
        if excludes := data.get("excludes"):
            project_config.excludes = excludes

        # reports
        if reports := data.get("reports"):
            for report_type, path in reports.items():
                path = self.resolve_path(path)
                project_config.reports.append(ReportConfig(report_type, path))

        # git
        if git := data.get("git"):
            project_config.git = GitConfig()
            project_config.git.repo = project_config.base_path
            project_config.git.branch = git.get("branch", project_config.git.branch)

        # composer
        if composer := data.get("composer"):
            project_config.composer = composer

        # pip
        if pip := data.get("pip"):
            project_config.pip = pip

        # npm
        if npm := data.get("npm"):
            project_config.npm = npm

        # trends
        if history_path := data.get("trends"):
            project_config.history_path = self.resolve_path(history_path)

        # code smells
        code_smells = data.get("code_smells", True)
        if not isinstance(code_smells, dict):
            if code_smells == False or code_smells is None:
                project_config.code_smells.disable_all()
        else:
            for code_smell, value in code_smells.items():
                project_config.code_smells.set(code_smell, value)

        return project_config



"""Python-specific AST parser implementation using tree-sitter"""

from typing import List, Optional

from tree_sitter import Node

from metripy.LangAnalyzer.Generic.Ast.AstParser import AstParser


class PythonAstParser(AstParser):
    """Python-specific implementation of AST parser"""

    def __init__(self):
        super().__init__("python")

    def get_import_nodes(self) -> List[Node]:
        """Get all import statements"""
        imports = self.find_nodes_by_type("import_statement")
        imports.extend(self.find_nodes_by_type("import_from_statement"))
        return imports

    def get_function_nodes(self) -> List[Node]:
        """Get all function definitions"""
        return self.find_nodes_by_type("function_definition")

    def get_class_nodes(self) -> List[Node]:
        """Get all class definitions"""
        return self.find_nodes_by_type("class_definition")

    def get_variable_assignment_nodes(self) -> List[Node]:
        """Get all assignment statements"""
        return self.find_nodes_by_type("assignment")

    def get_identifier_nodes(self, context: str = "usage") -> List[Node]:
        """Get identifier nodes based on context"""
        return self.find_nodes_by_type("identifier")

    def extract_function_name(self, node: Node) -> Optional[str]:
        """Extract function name from function_definition node"""
        for child in node.children:
            if child.type == "identifier":
                return self.get_node_text(child)
        return None

    def extract_class_name(self, node: Node) -> Optional[str]:
        """Extract class name from class_definition node"""
        for child in node.children:
            if child.type == "identifier":
                return self.get_node_text(child)
        return None

    def extract_variable_name(self, node: Node) -> Optional[str]:
        """Extract variable name from assignment node"""
        # In Python, assignment structure is: left = right
        # The left side contains the identifier
        for child in node.children:
            if child.type == "identifier":
                return self.get_node_text(child)
            elif child.type in ("pattern_list", "tuple_pattern"):
                # Handle tuple unpacking: a, b = ...
                names = []
                for grandchild in child.children:
                    if grandchild.type == "identifier":
                        names.append(self.get_node_text(grandchild))
                return ", ".join(names) if names else None
        return None

    def extract_import_name(self, node: Node) -> Optional[str]:
        """Extract imported name from import node"""
        name = self.extract_import_qualified_name(node)
        if name:
            return name.split(".")[-1]
        return None

    def extract_import_qualified_name(self, node: Node) -> Optional[str]:
        """Extract imported module/name from import node"""
        if node.type == "import_statement":
            # import foo or import foo as bar
            for child in node.children:
                if child.type == "dotted_name":
                    return self.get_node_text(child)
                elif child.type == "aliased_import":
                    # Get the alias name
                    for grandchild in child.children:
                        if (
                            grandchild.type == "identifier"
                            and grandchild.next_sibling is None
                        ):
                            return self.get_node_text(grandchild)
                    # If no alias, get the original name
                    for grandchild in child.children:
                        if grandchild.type == "dotted_name":
                            return self.get_node_text(grandchild)

        elif node.type == "import_from_statement":
            # from foo import bar or from foo import bar as baz
            imported_names = []
            for child in node.children:
                if (
                    child.type == "dotted_name"
                    and child.prev_sibling
                    and self.get_node_text(child.prev_sibling) == "import"
                ):
                    imported_names.append(self.get_node_text(child))
                elif child.type == "aliased_import":
                    for grandchild in child.children:
                        if (
                            grandchild.type == "identifier"
                            and grandchild.next_sibling is None
                        ):
                            imported_names.append(self.get_node_text(grandchild))
                            break
                    else:
                        for grandchild in child.children:
                            if grandchild.type == "identifier":
                                imported_names.append(self.get_node_text(grandchild))
                                break
            return ", ".join(imported_names) if imported_names else None

        return None

    def get_class_methods(self, node: Node) -> List[Node]:
        methods: List[Node] = []
        for child in self.walk_tree(node):
            if child.type == "function_definition":
                methods.append(child)
        return methods

    def get_function_parameters(self, function_node: Node) -> List[str]:
        """Extract parameter names from function definition"""
        params = []
        for child in function_node.children:
            if child.type == "parameters":
                for param_node in child.children:
                    if param_node.type == "identifier":
                        params.append(self.get_node_text(param_node))
                    elif param_node.type in ("typed_parameter", "default_parameter"):
                        # Handle typed and default parameters
                        for grandchild in param_node.children:
                            if grandchild.type == "identifier":
                                params.append(self.get_node_text(grandchild))
                                break
        return params

    def get_function_attributes(self, function_node: Node) -> List[str]:
        """Get names of variables used in the function"""
        attributes = []
        for child in self.walk_tree(function_node):
            if child.type == "attribute" and not child.parent.type == "call":
                attributes.append(self.get_node_text(child))
        return attributes

    def _get_function_calls(self, function_node: Node) -> List[str]:
        """Get names of function variables"""
        attributes = []
        for child in self.walk_tree(function_node):
            if child.type == "attribute" and child.parent.type == "call":
                attributes.append(self.get_node_text(child))
        return attributes

    def get_function_self_calls(self, function_node: Node) -> List[str]:
        """Get names class methods called in the function"""
        calls = self._get_function_calls(function_node)
        self_method_calls = []
        for call in calls:
            parts = call.split(".")
            if parts[0] == "self" and len(parts) == 2:
                self_method_calls.append(parts[1])
        return self_method_calls

    def get_operator_types(self) -> List[str]:
        raise NotImplementedError

    def get_operand_types(self) -> List[str]:
        raise NotImplementedError



"""Generic detector for unused imports across languages"""

from typing import List, Set

from metripy.LangAnalyzer.Generic.CodeSmell.CodeSmell import (
    CodeSmell,
    CodeSmellSeverity,
    CodeSmellType,
)
from metripy.LangAnalyzer.Generic.CodeSmell.Detectors.BaseDetector import (
    BaseCodeSmellDetector,
)


class UnusedImportsDetector(BaseCodeSmellDetector):
    """Detects unused imports in any language"""

    def detect(self) -> List[CodeSmell]:
        """Detect unused imports"""
        import_nodes = self.parser.get_import_nodes()
        if not import_nodes:
            return []

        # Get all used identifiers in the code
        used_identifiers = self._get_used_identifiers()

        for import_node in import_nodes:
            import_name = self.parser.extract_import_name(import_node)
            if not import_name:
                continue

            # Handle multiple imports in one statement
            import_names = [name.strip() for name in import_name.split(",")]

            for name in import_names:
                if name and name not in used_identifiers:
                    # Check if it's a module import and any attributes are used
                    is_used = any(
                        used.startswith(f"{name}.") for used in used_identifiers
                    )

                    if not is_used:
                        self.smells.append(
                            CodeSmell(
                                smell_type=CodeSmellType.UNUSED_IMPORT,
                                severity=CodeSmellSeverity.MINOR,
                                filename=self.filename,
                                line_number=self.get_line_number(import_node),
                                column=self.get_column(import_node),
                                code_line=self.parser.get_code_line(import_node),
                                message=f"Unused import '{name}'",
                                symbol=name,
                            )
                        )

        return self.smells

    def _get_used_identifiers(self) -> Set[str]:
        """Get all identifiers used in the code"""
        used = set()

        # Get all identifier nodes that are not in import statements
        all_identifiers = self.parser.get_identifier_nodes("usage")
        import_nodes = self.parser.get_import_nodes()

        # Build set of import node ranges to exclude
        import_ranges = set()
        for import_node in import_nodes:
            for node in self.parser.walk_tree(import_node):
                import_ranges.add((node.start_byte, node.end_byte))

        # Add identifiers that are not part of import statements
        for identifier in all_identifiers:
            node_range = (identifier.start_byte, identifier.end_byte)
            if node_range not in import_ranges:
                text = self.parser.get_node_text(identifier)
                if text:
                    used.add(text)

        return used



from metripy.Application.Config.CodeSmellConfig import CodeSmellConfig
from metripy.LangAnalyzer.Generic.CodeSmell.GenericCodeSmellDetector import (
    GenericCodeSmellDetector,
)
from metripy.LangAnalyzer.Php.CodeSmell.PhpCodeSmellDetector import PhpCodeSmellDetector
from metripy.LangAnalyzer.Python.CodeSmell.PythonCodeSmellDetector import (
    PythonCodeSmellDetector,
)
from metripy.LangAnalyzer.Typescript.CodeSmell.TypescriptCodeSmellDetector import (
    TypescriptCodeSmellDetector,
)


class CodeSmellDetectorFactory:
    _DETECTORS = {
        "Python": PythonCodeSmellDetector,
        "PHP": PhpCodeSmellDetector,
        "Typescript": TypescriptCodeSmellDetector,
    }

    @staticmethod
    def get_code_smell_detector(
        language: str, config: CodeSmellConfig
    ) -> GenericCodeSmellDetector:
        try:
            return CodeSmellDetectorFactory._DETECTORS[language](config)
        except KeyError:
            raise ValueError(f"No code smell detector found for language: {language}")



from metripy.Metric.ProjectMetrics import ProjectMetrics
from metripy.Report.Json.AbstractJsonReporter import AbstractJsonReporter


class GitJsonReporter(AbstractJsonReporter):
    def generate(self, metrics: ProjectMetrics) -> None:
        if not metrics.git_metrics:
            self.output.writeln(
                "<error>Wants git json report, but no git metrics</error>"
            )
            return

        data = {
            "commits_per_month": metrics.git_metrics.get_commit_stats_per_month(),
            "churn_data": metrics.git_metrics.get_churn_per_month(),
            "possible_silos": metrics.git_metrics.get_silos_list()[:10],
            "top_contributors": metrics.git_metrics.get_contributors_list()[:10],
            "top_hotspots": metrics.git_metrics.get_hotspots_list()[:10],
        }

        self.put_data(data)
        self.output.writeln(
            f"<success>Create git json report in {self.config.path}</success>"
        )



import json

from metripy.Metric.FileTree.FileTreeParser import FileTreeParser
from metripy.Metric.ProjectMetrics import ProjectMetrics
from metripy.Report.Html.PageRenderer import PageRenderer


class FilesPageRenderer(PageRenderer):
    def __init__(self, template_dir: str, output_dir: str, project_name: str):
        super().__init__(template_dir, output_dir, project_name)

    def render(self, metrics: ProjectMetrics):
        file_names = []
        file_details = {}
        for file_metrics in metrics.file_metrics:
            file_name = file_metrics.full_name
            file_details[file_name] = file_metrics.to_dict()
            file_names.append(file_name)

        filetree = FileTreeParser.parse(file_names, shorten=True)

        self.render_template(
            "files.html",
            {
                "filetree": json.dumps(filetree.to_dict()),
                "file_details": json.dumps(file_details),
            },
        )



from typing import Self

from metripy.LangAnalyzer.Generic.CodeSmell.CodeSmell import CodeSmell
from metripy.Metric.Code.Segmentor import Segmentor
from metripy.Metric.Trend.FileTrendMetric import FileTrendMetric
from metripy.Tree.ClassNode import ClassNode
from metripy.Tree.FunctionNode import FunctionNode


class FileMetrics:
    def __init__(
        self,
        full_name: str,
        loc: int,
        totalCc: int,
        avgCcPerFunction: float,
        maintainabilityIndex: float,
        avgLocPerFunction: float,
        class_nodes: list[ClassNode],
        function_nodes: list[FunctionNode],
        import_name: str | None,
        imports: list[str] | None,
        code_smells: list[CodeSmell],
        total_cog_complexity: int,
        avg_cog_complexity_per_function: float,
        avg_lcom4_per_class: float,
    ):
        self.full_name = full_name
        self.loc = loc
        self.totalCc = totalCc
        self.avgCcPerFunction = avgCcPerFunction
        self.maintainabilityIndex = maintainabilityIndex
        self.avgLocPerFunction = avgLocPerFunction
        self.class_nodes = class_nodes
        self.function_nodes = function_nodes
        self.trend: FileTrendMetric | None = None
        self.import_name = import_name
        self.imports: list[str] | None = imports
        self.imported_by: list[str] | None = None
        self.afferent_coupling: int = 0
        self.efferent_coupling: int = 0
        self.instability: float = 0
        self.code_smells: list[CodeSmell] = code_smells
        self.total_cog_complexity = total_cog_complexity
        self.avg_cog_complexity_per_function = avg_cog_complexity_per_function
        self.avg_lcom4_per_class = avg_lcom4_per_class

    def to_dict(self) -> dict:
        return {
            "full_name": self.full_name,
            "loc": self.loc,
            "loc_segment": Segmentor.get_loc_segment(self.loc),
            "totalCc": self.totalCc,
            "complexity_segment": Segmentor.get_complexity_segment(
                self.avgCcPerFunction
            ),
            "avgCcPerFunction": self.avgCcPerFunction,
            "maintainabilityIndex": round(self.maintainabilityIndex, 2),
            "maintainability_segment": Segmentor.get_maintainability_segment(
                self.maintainabilityIndex
            ),
            "avgLocPerFunction": self.avgLocPerFunction,
            "method_size_segment": Segmentor.get_method_size_segment(
                self.avgLocPerFunction
            ),
            "class_nodes": [node.to_dict() for node in self.class_nodes],
            "function_nodes": [node.to_dict() for node in self.function_nodes],
            "import_name": self.import_name,
            "imports": self.imports,
            "afferent_coupling": self.afferent_coupling,
            "efferent_coupling": self.efferent_coupling,
            "instability": round(self.instability, 2),
            "total_cog_complexity": self.total_cog_complexity,
            "avg_cog_complexity_per_function": self.avg_cog_complexity_per_function,
            "cognitive_complexity_segment": Segmentor.get_complexity_segment(
                self.avg_cog_complexity_per_function
            ),
            "avg_lcom4_per_class": self.avg_lcom4_per_class,
            "lcom4_segment": Segmentor.get_lcom4_segment(self.avg_lcom4_per_class),
        }

    @staticmethod
    def from_dict(data: dict) -> Self:
        metrics = FileMetrics(
            full_name=data["full_name"],
            loc=data["loc"],
            totalCc=data["totalCc"],
            avgCcPerFunction=data["avgCcPerFunction"],
            maintainabilityIndex=data["maintainabilityIndex"],
            avgLocPerFunction=data["avgLocPerFunction"],
            class_nodes=[ClassNode.from_dict(d) for d in data["class_nodes"]],
            function_nodes=[FunctionNode.from_dict(d) for d in data["function_nodes"]],
            import_name=data.get("import_name"),
            imports=data.get("imports"),
            code_smells=[CodeSmell.from_dict(d) for d in data.get("code_smells", [])],
            total_cog_complexity=data.get("total_cog_complexity", 0),
            avg_cog_complexity_per_function=data.get(
                "avg_cog_complexity_per_function", 0
            ),
            avg_lcom4_per_class=data.get("avg_lcom4_per_class", 0),
        )
        metrics.afferent_coupling = data.get("afferent_coupling", 0)
        metrics.efferent_coupling = data.get("efferent_coupling", 0)
        metrics.instability = data.get("instability", 0)
        return metrics



import requests

from metripy.Dependency.Dependency import Dependency


class PyPi:
    def get_info(self, dependency: Dependency) -> Dependency | None:
        uri = f"https://pypi.org/pypi/{dependency.name}/json"
        x = requests.get(uri)
        data = x.json()

        info = data.get("info", {})
        releases = data.get("releases", {})

        if not info:
            print(f"Package '{dependency.name}' has no info section")
            return dependency

        dependency.type = "pip"
        dependency.description = info.get("summary")
        dependency.repository = info.get("project_url") or info.get("home_page")
        if info.get("license"):
            dependency.license = [info.get("license")]
        else:
            dependency.license = []
        dependency.homepage = info.get("home_page")

        # PyPI doesn't provide GitHub stars or download counts directly
        dependency.github_stars = "??"
        dependency.downloads_total = "??"
        dependency.downloads_monthly = "??"
        dependency.downloads_daily = "??"

        # Determine latest version
        latest_version = info.get("version")
        dependency.latest = latest_version

        # Compare with current version
        if dependency.version == latest_version:
            dependency.status = "latest"
        else:
            dependency.status = "outdated"

        # Get distribution URL (e.g., wheel or sdist)
        if latest_version in releases:
            release_files = releases[latest_version]
            if release_files:
                dependency.zip = release_files[0].get("url")

        return dependency



from metripy.Metric.Code.FileMetrics import FileMetrics
from metripy.Report.Cli.AbstractFormatter import AbstractFormatter


class CouplingMetricsFormatter(AbstractFormatter):
    def format(self, file_metrics: list[FileMetrics]):
        return f"""
========= Coupling Metrics =========
Total Modules: {self._get_total_modules(file_metrics)}
Total Dependencies: {self._get_total_dependencies(file_metrics)}
Max Coupling: {self._get_max_coupling(file_metrics)}
Average Instability: {round(self._get_average_instability(file_metrics), 2)}
{self._format_instability_bar(file_metrics)}
"""

    def _get_total_modules(self, file_metrics: list[FileMetrics]) -> int:
        return len(file_metrics)

    def _get_total_dependencies(self, file_metrics: list[FileMetrics]) -> int:
        return sum(len(f.imports) for f in file_metrics)

    def _get_average_instability(self, file_metrics: list[FileMetrics]) -> float:
        return sum(f.instability for f in file_metrics) / len(file_metrics)

    def _get_max_coupling(self, file_metrics: list[FileMetrics]) -> int:
        return max(f.afferent_coupling + f.efferent_coupling for f in file_metrics)

    def _format_instability_bar(self, file_metrics: list[FileMetrics]) -> str:
        stable = len([f.instability for f in file_metrics if f.instability <= 0.3])
        moderate = len(
            [
                f.instability
                for f in file_metrics
                if f.instability > 0.3 and f.instability <= 0.7
            ]
        )
        unstable = len([f.instability for f in file_metrics if f.instability > 0.7])

        colors = [self.COLORS["good"], self.COLORS["ok"], self.COLORS["critical"]]
        return f"""
{self.colored_stacked_bar([stable, moderate, unstable], colors, width=100)}
{self.colored_stacked_info([stable, moderate, unstable], ["Stable", "Neutral", "Unstable"], colors, width=100)}
"""



from metripy.LangAnalyzer.Generic.Ast.AstParser import AstParser
from metripy.LangAnalyzer.Php.Ast.PhpAstParser import PhpAstParser
from metripy.LangAnalyzer.Python.Ast.PythonAstParser import PythonAstParser


class AstParserFactory:
    _PARSERS = {
        "Python": PythonAstParser(),
        "PHP": PhpAstParser(),
        "Typescript": None,
    }

    @staticmethod
    def get_ast_parser(language: str) -> AstParser:
        try:
            return AstParserFactory._PARSERS[language]
        except KeyError:
            raise ValueError(f"No ast parser found for language: {language}")



from metripy.LangAnalyzer.Generic.Metrics.GenericLcom4Analyzer import (
    GenericLcom4Analyzer,
)


class PhpLcom4Analyzer(GenericLcom4Analyzer):
    def get_methods_to_ignore(self) -> list[str]:
        return [
            "__construct",
            "__destruct",
            "__clone",
            "__sleep",
            "__wakeup",
            "__toString",
            "__invoke",
            "__set",
            "__get",
            "__set_state",
            "__clone",
            "__debugInfo",
        ]



from typing import Self

from metripy.Metric.Code.Segmentor import Segmentor
from metripy.Metric.Trend.ClassTrendMetric import ClassTrendMetric
from metripy.Tree.FunctionNode import FunctionNode


class ClassNode:
    def __init__(
        self,
        full_name: str,
        name: str,
        lineno: int,
        line_end: int,
        real_complexity: int,
    ):
        self.full_name = full_name
        self.name = name
        self.lineno = lineno
        self.line_end = line_end
        self.real_complexity = real_complexity
        self.lcom4: int | None = None
        self.functions: list[FunctionNode] = []

        self.trend: ClassTrendMetric | None = None

    def to_dict(self) -> dict:
        """Convert ClassNode to a dictionary for JSON serialization."""
        return {
            "full_name": self.full_name,
            "name": self.name,
            "lineno": self.lineno,
            "line_end": self.line_end,
            "real_complexity": self.real_complexity,
            "lcom4": self.lcom4,
            "complexity_segment": Segmentor.get_complexity_segment(
                self.real_complexity
            ),
            "functions": [func.to_dict() for func in self.functions],
        }

    def __dict__(self) -> dict:
        return self.to_dict()

    @staticmethod
    def from_dict(data: dict) -> Self:
        node = ClassNode(
            full_name=data["full_name"],
            name=data["name"],
            lineno=data["lineno"],
            line_end=data.get("line_end", 0),
            real_complexity=data["real_complexity"],
        )
        node.lcom4 = data.get("lcom4", None)
        node.functions = [FunctionNode.from_dict(d) for d in data["functions"]]
        return node



class FunctionTrendMetric:
    def __init__(
        self,
        historical_loc: int,
        loc: int,
        historical_complexity: int,
        complexity: int,
        historical_maintainability_index: float,
        maintainability_index: float,
    ):
        self.historical_loc = historical_loc
        self.loc_delta = loc - historical_loc
        self.historical_complexity = historical_complexity
        self.complexity_delta = complexity - historical_complexity
        self.historical_maintainability_index = historical_maintainability_index
        self.maintainability_index_delta = (
            maintainability_index - historical_maintainability_index
        )

    def to_dict(self) -> dict:
        return {
            "historical_loc": self.historical_loc,
            "loc_delta": self.loc_delta,
            "historical_complexity": self.historical_complexity,
            "complexity_delta": self.complexity_delta,
            "historical_maintainability_index": self.historical_maintainability_index,
            "maintainability_index_delta": self.maintainability_index_delta,
        }



from metripy.LangAnalyzer.Generic.Metrics.GenericImportsAnalyzer import (
    GenericImportsAnalyzer,
)
from metripy.LangAnalyzer.Generic.Ast.AstParser import AstParser


class PhpImportsAnalyzer(GenericImportsAnalyzer):
    def extract_import_name(self, filename: str, parser: AstParser) -> str:
        return parser.get_fqcn(filename)

    def extract_imports(self, filename: str, parser: AstParser) -> list[str]:
        imports = []
        for node in parser.get_import_nodes():
            imports.append(parser.extract_import_qualified_name(node))
        return imports



"""Generic base for naming convention detection"""

import re
from abc import abstractmethod
from typing import Dict, List

from metripy.Application.Config.CodeSmellConfig import CodeSmellConfig
from metripy.LangAnalyzer.Generic.CodeSmell.CodeSmell import CodeSmell
from metripy.LangAnalyzer.Generic.CodeSmell.Detectors.BaseDetector import (
    BaseCodeSmellDetector,
)


class NamingConventionDetector(BaseCodeSmellDetector):
    """Base class for naming convention detection with language-specific rules"""

    PATTERN_SNAKE_CASE = r"^[a-z][a-z0-9_]*$"
    PATTERN_CAMEL_CASE = r"^[a-z][a-zA-Z0-9]*$"
    PATTERN_PASCAL_CASE = r"^[A-Z][a-zA-Z0-9]*$"
    PATTERN_SCREAMING_SNAKE_CASE = r"^[A-Z][A-Z0-9_]*$"

    def __init__(self, filename: str, parser, config: CodeSmellConfig):
        super().__init__(filename, parser, config)
        # Define naming rules for different elements
        # Override these in language-specific implementations
        self.naming_rules: Dict[str, Dict] = {
            k: v
            for k, v in self._get_naming_rules().items()
            if self.config.is_enabled(v["smell_type"].value)
        }

    @abstractmethod
    def _get_naming_rules(self) -> Dict[str, Dict]:
        """
        Define naming rules for the language.

        Returns dict with structure:
        {
            'class': {
                'pattern': r'^[A-Z][a-zA-Z0-9]*$',  # regex pattern
                'description': 'PascalCase',
                'smell_type': CodeSmellType.PASCAL_CASE_VIOLATION_CLASS,
                'severity': CodeSmellSeverity.MINOR,
                'excludes': []  # patterns to exclude (e.g., magic methods)
            },
            ...
        }
        """
        pass

    def detect(self) -> List[CodeSmell]:
        """Detect naming convention violations"""
        # Check class names
        if "class" in self.naming_rules:
            self._check_classes()

        # Check function names
        if "function" in self.naming_rules:
            self._check_functions()

        # Check variable names
        if "variable" in self.naming_rules:
            self._check_variables()

        return self.smells

    def _check_classes(self) -> None:
        """Check class naming conventions"""
        rule = self.naming_rules["class"]
        class_nodes = self.parser.get_class_nodes()

        for class_node in class_nodes:
            class_name = self.parser.extract_class_name(class_node)
            if not class_name:
                continue

            # Check exclusions
            if self._should_exclude(class_name, rule.get("excludes", [])):
                continue

            if not re.match(rule["pattern"], class_name):
                self.smells.append(
                    CodeSmell(
                        smell_type=rule["smell_type"],
                        severity=rule["severity"],
                        filename=self.filename,
                        line_number=self.get_line_number(class_node),
                        column=self.get_column(class_node),
                        message=f"Class name '{class_name}' should use {rule['description']}",
                        code_line=self.parser.get_code_line(class_node),
                        symbol=class_name,
                    )
                )

    def _check_functions(self) -> None:
        """Check function naming conventions"""
        rule = self.naming_rules["function"]
        function_nodes = self.parser.get_function_nodes()

        for func_node in function_nodes:
            func_name = self.parser.extract_function_name(func_node)
            if not func_name:
                continue

            # Check exclusions
            if self._should_exclude(func_name, rule.get("excludes", [])):
                continue

            if not re.match(rule["pattern"], func_name):
                self.smells.append(
                    CodeSmell(
                        smell_type=rule["smell_type"],
                        severity=rule["severity"],
                        filename=self.filename,
                        line_number=self.get_line_number(func_node),
                        column=self.get_column(func_node),
                        message=f"Function name '{func_name}' should use {rule['description']}",
                        code_line=self.parser.get_code_line(func_node),
                        symbol=func_name,
                    )
                )

    def _check_variables(self) -> None:
        """Check variable naming conventions"""
        rule = self.naming_rules["variable"]
        var_nodes = self.parser.get_variable_assignment_nodes()

        for var_node in var_nodes:
            var_name = self.parser.extract_variable_name(var_node)
            if not var_name:
                continue

            # Handle multiple variables in one statement
            var_names = [v.strip() for v in var_name.split(",")]

            for name in var_names:
                # Check exclusions
                if self._should_exclude(name, rule.get("excludes", [])):
                    continue

                # Check for constants (all uppercase)
                if "constant" in self.naming_rules and name.isupper():
                    const_rule = self.naming_rules["constant"]
                    if not re.match(const_rule["pattern"], name):
                        self.smells.append(
                            CodeSmell(
                                smell_type=const_rule["smell_type"],
                                severity=const_rule["severity"],
                                filename=self.filename,
                                line_number=self.get_line_number(var_node),
                                column=self.get_column(var_node),
                                message=f"Constant '{name}' should use {const_rule['description']}",
                                code_line=self.parser.get_code_line(var_node),
                                symbol=name,
                            )
                        )
                elif not re.match(rule["pattern"], name):
                    self.smells.append(
                        CodeSmell(
                            smell_type=rule["smell_type"],
                            severity=rule["severity"],
                            filename=self.filename,
                            line_number=self.get_line_number(var_node),
                            column=self.get_column(var_node),
                            message=f"Variable name '{name}' should use {rule['description']}",
                            code_line=self.parser.get_code_line(var_node),
                            symbol=name,
                        )
                    )

    def _should_exclude(self, name: str, exclusions: List[str]) -> bool:
        """Check if name should be excluded from checking"""
        for exclusion in exclusions:
            if callable(exclusion):
                if exclusion(name):
                    return True
            elif re.match(exclusion, name):
                return True
        return False



class Segmentor:
    @staticmethod
    def get_loc_segment(loc: int) -> str:
        if loc <= 200:
            return "good"
        elif loc <= 500:
            return "ok"
        elif loc <= 1000:
            return "warning"
        else:
            return "critical"

    @staticmethod
    def get_complexity_segment(complexity: float) -> str:
        if complexity <= 5:
            return "good"
        elif complexity <= 10:
            return "ok"
        elif complexity <= 20:
            return "warning"
        else:
            return "critical"

    @staticmethod
    def get_maintainability_segment(maintainability: float) -> str:
        if maintainability > 80:
            return "good"
        elif maintainability > 60:
            return "ok"
        elif maintainability > 40:
            return "warning"
        else:
            return "critical"

    @staticmethod
    def get_method_size_segment(method_size: float) -> str:
        if method_size <= 15:
            return "good"
        elif method_size <= 30:
            return "ok"
        elif method_size <= 50:
            return "warning"
        else:
            return "critical"

    @staticmethod
    def get_lcom4_segment(lcom4: float) -> str:
        # only 1 good, 0 is useless and above 1 is bad
        if lcom4 == 1:
            return "good"
        elif lcom4 == 0:
            return "warning"
        else:
            return "critical"



from metripy.Application.Config.CodeSmellConfig import CodeSmellConfig
from metripy.Application.Config.GitConfig import GitConfig
from metripy.Application.Config.ReportConfig import ReportConfig


class ProjectConfig:
    def __init__(self, name: str):
        self.name: str = name
        self.base_path: str = "./"
        self.includes: list[str] = []
        self.excludes: list[str] = []
        self.extensions: list[str] = []
        self.git: GitConfig | None = None
        self.composer: bool = False
        self.pip: bool = False
        self.npm: bool = False
        self.reports: list[ReportConfig] = []
        self.history_path: str | None = None
        self.code_smells: CodeSmellConfig = CodeSmellConfig()

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "base_path": self.base_path,
            "includes": self.includes,
            "excludes": self.excludes,
            "extensions": self.extensions,
            "composer": self.composer,
            "pip": self.pip,
            "npm": self.npm,
            "git": self.git.to_dict() if self.git else None,
            "reports": [report.to_dict() for report in self.reports],
            "history_path": self.history_path,
            "code_smells": self.code_smells.to_dict(),
        }

    @staticmethod
    def str_to_bool(value):
        if isinstance(value, bool):
            return value
        return str(value).lower() in ("true", "1", "yes")

    def set(self, keys: list[str], value: any) -> None:
        if len(keys) == 0:
            return
        primary_key = keys[0]
        # single value
        if primary_key == "base_path":
            self.base_path = value
        elif primary_key == "pip":
            self.pip = self.str_to_bool(value)
        elif primary_key == "npm":
            self.npm = self.str_to_bool(value)
        elif primary_key == "composer":
            self.composer = self.str_to_bool(value)
        elif primary_key == "trends":
            self.history_path = value
        elif primary_key == "git":
            self.git = GitConfig()
            self.git.repo = self.base_path
            self.git.branch = value

        # list values
        elif primary_key == "includes":
            if value == "":
                self.includes = []
            else:
                self.includes.append(value)
        elif primary_key == "excludes":
            if value == "":
                self.excludes = []
            else:
                self.excludes.append(value)
        elif primary_key == "extensions":
            if value == "":
                self.extensions = []
            else:
                self.extensions.append(value)

        # dict values
        elif primary_key == "reports":
            if len(keys) == 1:
                return
            report_type = keys[1]
            report_path = value
            if value != "":
                self.reports.append(ReportConfig(report_type, report_path))
            else:
                report_config = next(
                    (rc for rc in self.reports if rc.type == report_type), None
                )
                if not report_config:
                    return
                self.reports.remove(report_config)



from metripy.LangAnalyzer.Generic.CodeSmell.CodeSmell import CodeSmellSeverity
from metripy.Metric.Code.FileMetrics import FileMetrics
from metripy.Report.Cli.AbstractFormatter import AbstractFormatter


class CodeSmellsFormatter(AbstractFormatter):
    def format(self, file_metrics: list[FileMetrics]):
        return f"""
========= Code Smells =========
{self._format_code_smells(file_metrics)}
"""

    def _format_code_smells(self, file_metrics: list[FileMetrics]) -> str:
        major = 0
        critical = 0
        minor = 0
        info = 0
        for f in file_metrics:
            for smell in f.code_smells:
                if smell.severity == CodeSmellSeverity.MAJOR:
                    major += 1
                if smell.severity == CodeSmellSeverity.CRITICAL:
                    critical += 1
                if smell.severity == CodeSmellSeverity.MINOR:
                    minor += 1
                if smell.severity == CodeSmellSeverity.INFO:
                    info += 1
        colors = [
            "34",
            self.COLORS["ok"],
            self.COLORS["warning"],
            self.COLORS["critical"],
        ]
        data = [info, minor, major, critical]
        labels = ["Info", "Minor", "Major", "Critical"]

        return f"""Total Code Smells: {sum(data)}

{self.colored_stacked_bar(data, colors, width=100)}
{self.colored_stacked_info(data, labels, colors, width=100)}
"""



import math
import re
from abc import ABC, abstractmethod
from collections import Counter


# TODO: radon only looks at unary, binary and boolean operators,
#  assign and function head
class GenericHalSteadAnalyzer(ABC):
    @abstractmethod
    def get_operators(self) -> list[str]:
        pass

    def calculate_halstead_metrics(self, code: str):
        # Tokenize the code
        tokens = re.findall(r"\b\w+\b|[^\s\w]", code)

        # Count operators and operands
        operator_counts = Counter()
        operand_counts = Counter()

        for token in tokens:
            if token in self.get_operators():
                operator_counts[token] += 1
            elif re.match(r"\b\w+\b", token):
                operand_counts[token] += 1

        # Halstead metrics
        n1 = len(operator_counts)  # distinct operators
        n2 = len(operand_counts)  # distinct operands
        N1 = sum(operator_counts.values())  # total operators
        N2 = sum(operand_counts.values())  # total operands

        vocabulary = n1 + n2
        length = N1 + N2
        volume = length * math.log2(vocabulary) if vocabulary > 0 else 0
        difficulty = (n1 / 2) * (N2 / n2) if n2 > 0 else 0
        effort = difficulty * volume

        calculated_length = 0
        if n1 > 0:
            calculated_length += n1 * math.log2(n1)
        if n2 > 0:
            calculated_length += n2 * math.log2(n2)

        bugs = (effort ** (2 / 3)) / 3000 if effort > 0 else 0
        time = effort / 18 if effort > 0 else 0

        return {
            "n1": n1,  # distinct operators
            "n2": n2,  # distinct operands
            "N1": N1,  # total operators
            "N2": N2,  # total operands
            "vocabulary": vocabulary,
            "length": length,
            "volume": volume,
            "difficulty": difficulty,
            "effort": effort,
            "calculated_length": calculated_length,
            "bugs": bugs,
            "time": time,
        }



from metripy.LangAnalyzer.Generic.Metrics.GenericHalSteadAnalyzer import (
    GenericHalSteadAnalyzer,
)


class PhpHalSteadAnalyzer(GenericHalSteadAnalyzer):
    def get_operators(self) -> list[str]:
        return [
            "+",
            "-",
            "*",
            "/",
            "%",
            "++",
            "--",
            "=",
            "==",
            "!=",
            "===",
            "!==",
            "<",
            ">",
            "<=",
            ">=",
            "&&",
            "||",
            "!",
            ".",
            "->",
            "=>",
            "::",
            "?",
            ":",
            "&",
            "|",
            "^",
            "~",
            "<<",
            ">>",
        ]



"""Generic detector for unused functions across languages"""

from typing import List, Set

from metripy.LangAnalyzer.Generic.CodeSmell.CodeSmell import (
    CodeSmell,
    CodeSmellSeverity,
    CodeSmellType,
)
from metripy.LangAnalyzer.Generic.CodeSmell.Detectors.BaseDetector import (
    BaseCodeSmellDetector,
)


class UnusedFunctionsDetector(BaseCodeSmellDetector):
    """Detects unused private functions in any language"""

    def detect(self) -> List[CodeSmell]:
        """Detect unused private functions"""
        function_nodes = self.parser.get_function_nodes()
        if not function_nodes:
            return []

        # Track private function definitions
        private_functions = {}  # name -> node
        called_functions = self._get_called_functions()

        for func_node in function_nodes:
            func_name = self.parser.extract_function_name(func_node)
            if not func_name:
                continue

            # Check if it's a private function (starts with _ in Python, TypeScript convention)
            if self._is_private_function(func_name):
                private_functions[func_name] = func_node

        # Find unused private functions
        for func_name, func_node in private_functions.items():
            if func_name not in called_functions:
                self.smells.append(
                    CodeSmell(
                        smell_type=CodeSmellType.UNUSED_FUNCTION,
                        severity=CodeSmellSeverity.MINOR,
                        filename=self.filename,
                        line_number=self.get_line_number(func_node),
                        column=self.get_column(func_node),
                        end_line=self.get_end_line(func_node),
                        code_line=self.parser.get_code_line(func_node),
                        message=f"Unused private function '{func_name}'",
                        symbol=func_name,
                    )
                )

        return self.smells

    def _is_private_function(self, name: str) -> bool:
        """Check if function name indicates it's private"""
        # Python/TypeScript convention: starts with _ but not __
        if name.startswith("_") and not name.startswith("__"):
            return True
        # TypeScript might also use # for private
        if name.startswith("#"):
            return True
        return False

    def _get_called_functions(self) -> Set[str]:
        """Get all function names that are called in the code"""
        called = set()

        # Get all identifiers that could be function calls
        identifiers = self.parser.get_identifier_nodes("usage")

        for identifier in identifiers:
            name = self.parser.get_node_text(identifier)
            if name:
                called.add(name)

        return called



import math

import lizard

from metripy.Application.Config.ProjectConfig import ProjectConfig
from metripy.LangAnalyzer.AbstractLangAnalyzer import AbstractLangAnalyzer
from metripy.LangAnalyzer.Typescript.TypescriptAstParser import TypescriptAstParser
from metripy.LangAnalyzer.Typescript.TypescriptBasicComplexityAnalyzer import (
    TypescriptBasicComplexityAnalzyer,
)
from metripy.LangAnalyzer.Typescript.TypescriptBasicLocAnalyzer import (
    TypescriptBasicLocAnalyzer,
)
from metripy.LangAnalyzer.Typescript.TypescriptHalSteadAnalyzer import (
    TypeScriptHalSteadAnalyzer,
)
from metripy.Tree.ClassNode import ClassNode
from metripy.Tree.FunctionNode import FunctionNode
from metripy.Tree.ModuleNode import ModuleNode


class TypescriptAnalyzer(AbstractLangAnalyzer):

    def __init__(self, project_config: ProjectConfig):
        super().__init__(project_config)
        self.ast_parser = TypescriptAstParser()
        self.halstead_analyzer = TypeScriptHalSteadAnalyzer()
        self.basic_complexity_analyzer = TypescriptBasicComplexityAnalzyer()

    def get_lang_name(self) -> str:
        return "Typescript"

    def get_supported_extensions(self) -> tuple[str]:
        return (".ts", ".js")

    def analyze(self, code: str, filename: str) -> None:
        structure = self.ast_parser.extract_structure(code)

        lizard_result = lizard.analyze_file(filename)
        complexity_data = {
            func.name: {
                "complexity": func.cyclomatic_complexity,
                "start_line": func.start_line,
                "end_line": func.end_line,
            }
            for func in lizard_result.function_list
        }

        classes: dict[str, ClassNode] = {}
        functions: dict[str, FunctionNode] = {}

        for func_name in structure["functions"]:
            full_name = self.full_name(filename, func_name)
            try:
                function_complexity_data = complexity_data[func_name]
            except KeyError as e:
                function_complexity_data = (
                    self.basic_complexity_analyzer.get_complexity(code, func_name)
                )
                if function_complexity_data is None:
                    print(
                        f"error for function {full_name}: no lizard complexity data, basic analyzer also failed: '{e}'"
                    )
                    continue
            function_node = FunctionNode(
                full_name,
                func_name,
                function_complexity_data["start_line"],
                0,
                function_complexity_data["complexity"],
            )
            functions[full_name] = function_node

        for class_name, method_names in structure["classes"].items():
            full_name = self.full_name(filename, class_name)
            class_node = ClassNode(
                full_name,
                class_name,
                0,
                0,
                0.0,  # gets filled in later based on methods
            )
            classes[full_name] = class_node
            for func_name in method_names:
                full_name = self.full_name(filename, func_name, class_name)
                try:
                    function_complexity_data = complexity_data[func_name]
                except KeyError as e:
                    function_complexity_data = (
                        self.basic_complexity_analyzer.get_complexity(code, func_name)
                    )
                    if function_complexity_data is None:
                        print(
                            f"error for method {full_name}: no lizard complexity data, basic analyzer also failed: '{e}'"
                        )
                        continue
                function_node = FunctionNode(
                    full_name,
                    func_name,
                    function_complexity_data["start_line"],
                    function_complexity_data["end_line"],
                    function_complexity_data["complexity"],
                )
                class_node.functions.append(function_node)
                functions[full_name] = function_node

        # complexity of classes
        for class_node in classes.values():
            class_node.real_complexity = sum(
                [func.complexity for func in class_node.functions]
            )

        loc_data = TypescriptBasicLocAnalyzer.get_loc_metrics(code, filename)

        full_name = self.full_name(filename)
        module_node = ModuleNode(
            full_name,
            loc_data.get("lines", 0),
            loc_data.get("linesOfCode", 0),
            loc_data.get("logicalLinesOfCode", 0),
            loc_data.get("commentLines", 0),
            0,  # multi-line comments - not directly available
            loc_data.get("linesOfCode", 0) - loc_data.get("logicalLinesOfCode", 0),
            loc_data.get("commentLines", 0),
        )
        module_node.classes.extend(classes.values())
        module_node.functions.extend(functions.values())

        code_lines = code.split("\n")
        for func_name, function_node in functions.items():
            lines = code_lines[function_node.lineno : function_node.line_end]
            function_metrics = self.halstead_analyzer.calculate_halstead_metrics(
                "\n".join(lines)
            )
            function_node.h1 = function_metrics["n1"]
            function_node.h2 = function_metrics["n2"]
            function_node.N1 = function_metrics["N1"]
            function_node.N2 = function_metrics["N2"]
            function_node.vocabulary = function_metrics["vocabulary"]
            function_node.length = function_metrics["length"]
            function_node.volume = function_metrics["volume"]
            function_node.difficulty = function_metrics["difficulty"]
            function_node.effort = function_metrics["effort"]
            function_node.calculated_length = function_metrics["calculated_length"]
            function_node.bugs = function_metrics["bugs"]
            function_node.time = function_metrics["time"]
            function_node.calc_mi()

        maintainability_index = self._calculate_maintainability_index(
            functions.values(), module_node
        )
        module_node.maintainability_index = maintainability_index
        self.modules[full_name] = module_node

    def _calculate_maintainability_index(
        self, functions: list[FunctionNode], module_node: ModuleNode
    ) -> float:
        """Calculate maintainability index for PHP"""
        if not functions:
            return 100.0

        total_volume = sum(func.volume for func in functions)
        total_complexity = sum(func.complexity for func in functions)
        total_length = sum(func.length for func in functions)

        if total_volume == 0 or total_length == 0:
            return 100.0

        # PHP maintainability index calculation
        mi_base = max(
            (
                171
                - 5.2 * math.log(total_volume)
                - 0.23 * total_complexity
                - 16.2 * math.log(total_length)
            )
            * 100
            / 171,
            0,
        )

        # Comment weight
        comment_weight = 0
        if module_node.loc > 0:
            comment_ratio = module_node.single_comments / module_node.loc
            comment_weight = 50 * math.sin(math.sqrt(2.4 * comment_ratio))

        return mi_base + comment_weight

    def get_metrics(self):
        return super().get_metrics()



class FileTrendMetric:
    def __init__(
        self,
        historical_loc: int,
        loc: int,
        historical_totalCc: int,
        totalCc: int,
        historical_avgCcPerFunction: float,
        avgCcPerFunction: float,
        historical_maintainabilityIndex: float,
        maintainabilityIndex: float,
        historical_avgLocPerFunction: float,
        avgLocPerFunction: float,
        historical_avg_cog_complexity_per_function: float,
        avg_cog_complexity_per_function: float,
    ):
        self.historical_loc = historical_loc
        self.loc_delta = loc - historical_loc

        self.historical_totalCc = historical_totalCc
        self.totalCc_delta = totalCc - historical_totalCc

        self.historical_avgCcPerFunction = historical_avgCcPerFunction
        self.avgCcPerFunction_delta = avgCcPerFunction - historical_avgCcPerFunction

        self.maintainabilityIndex_delta = (
            maintainabilityIndex - historical_maintainabilityIndex
        )
        self.historical_maintainabilityIndex = historical_maintainabilityIndex

        self.avgLocPerFunction_delta = avgLocPerFunction - historical_avgLocPerFunction
        self.historical_avgLocPerFunction = historical_avgLocPerFunction

        self.historical_avg_cog_complexity_per_function = (
            historical_avg_cog_complexity_per_function
        )
        self.avg_cog_complexity_per_function_delta = (
            avg_cog_complexity_per_function - historical_avg_cog_complexity_per_function
        )

    def to_dict(self) -> dict:
        return {
            "historical_loc": round(self.historical_loc, 2),
            "loc_delta": round(self.loc_delta, 2),
            "historical_totalCc": round(self.historical_totalCc, 2),
            "totalCc_delta": round(self.totalCc_delta, 2),
            "historical_avgCcPerFunction": round(self.historical_avgCcPerFunction, 2),
            "avgCcPerFunction_delta": round(self.avgCcPerFunction_delta, 2),
            "historical_avg_cog_complexity_per_function": round(
                self.historical_avg_cog_complexity_per_function, 2
            ),
            "avg_cog_complexity_per_function_delta": round(
                self.avg_cog_complexity_per_function_delta, 2
            ),
            "historical_maintainabilityIndex": round(
                self.historical_maintainabilityIndex, 2
            ),
            "maintainabilityIndex_delta": round(self.maintainabilityIndex_delta, 2),
            "historical_avgLocPerFunction": round(self.historical_avgLocPerFunction, 2),
            "avgLocPerFunction_delta": round(self.avgLocPerFunction_delta, 2),
        }



class GitKnowledgeSilo:
    def __init__(self, file_path: str, owner: str, commits_count: int):
        self.file_path = file_path
        self.owner = owner
        self.commits_count = commits_count
        self.risk_level = self._calc_risk_level(commits_count)
        self.risk_label = self._calc_risk_label(self.risk_level)

    def _calc_risk_level(self, commits_count: int) -> str:
        if commits_count >= 15:
            return "high"
        elif commits_count >= 8:
            return "medium"
        else:
            return "low"

    def _calc_risk_label(self, risk_level: str) -> str:
        return risk_level.capitalize()

    def to_dict(self) -> dict[str, int]:
        return {
            "file_path": f"{self.file_path}",
            "owner": f"{self.owner}",
            "commits_count": f"{self.commits_count}",
            "risk_level": f"{self.risk_level}",
            "risk_label": f"{self.risk_label}",
        }



import ast
from abc import ABC, abstractmethod
from typing import List, Set

from metripy.LangAnalyzer.Generic.CodeSmell import CodeSmell


class BaseCodeSmellDetector(ABC):
    """Base class for all code smell detectors"""

    def __init__(self, filename: str, code: str, tree: ast.AST | None = None):
        self.filename = filename
        self.code = code
        self.tree = tree
        self.smells: List[CodeSmell] = []

    @abstractmethod
    def detect(self) -> List[CodeSmell]:
        """Detect code smells and return a list of CodeSmell objects"""
        pass

    def _get_used_names(self) -> Set[str]:
        """Get all names that are used (loaded) in the code"""
        used_names = set()

        if not self.tree:
            return used_names

        for node in ast.walk(self.tree):
            if isinstance(node, ast.Name) and isinstance(node.ctx, ast.Load):
                used_names.add(node.id)
            elif isinstance(node, ast.Attribute):
                # For attribute access like 'module.function', track 'module'
                if isinstance(node.value, ast.Name):
                    used_names.add(node.value.id)
                    used_names.add(f"{node.value.id}.{node.attr}")

        return used_names



import json

from metripy.Component.Output.CliOutput import CliOutput
from metripy.Metric.ProjectMetrics import ProjectMetrics


class JsonImporter:
    def __init__(self, output: CliOutput):
        self.output = output

    def import_data(self, path: str) -> ProjectMetrics:
        self.output.writeln(f"<info>Importing data from {path}...</info>")
        with open(path, "r") as file:
            data = json.load(file)
            project_metrics = ProjectMetrics.from_dict(data)
        self.output.writeln("<success>Data imported successfuly</success>")
        return project_metrics



class CodeSmellConfig:
    def __init__(self):
        self.config = {
            "unused_imports": True,
            "unused_functions": True,
            "too_many_parameters": True,
            "long_function": True,
            "snake_case_violation_variable": True,
            "snake_case_violation_function": True,
            "pascal_case_violation_class": True,
            "screaming_snake_case_violation_constant": True,
        }

    def disable_all(self) -> None:
        for param in self.config:
            self.config[param] = False

    def set(self, param: str, value: bool) -> None:
        if param in self.config.keys():
            self.config[param] = value

    def is_enabled(self, param: str) -> bool:
        return self.config.get(param, True)

    def to_dict(self) -> dict:
        return self.config



import re

import requests
from packaging import version

from metripy.Dependency.Dependency import Dependency


class Packegist:
    def get_info(self, dependency: Dependency) -> Dependency | None:
        if "/" not in dependency.name:
            return None
        [user, name] = dependency.name.split("/", 2)
        uri = f"https://packagist.org/packages/{user}/{name}.json"

        x = requests.get(uri)
        d = x.json()

        package_info = d.get("package", None)
        if package_info is None:
            print(f"package of {dependency.name} has no package info")
            return dependency

        dependency.type = package_info["type"]
        dependency.description = package_info["description"]
        dependency.repository = package_info["repository"]
        dependency.github_stars = package_info["github_stars"]
        dependency.downloads_total = package_info["downloads"]["total"]
        dependency.downloads_monthly = package_info["downloads"]["monthly"]
        dependency.downloads_daily = package_info["downloads"]["daily"]

        latest = version.parse("0.0.0")
        versions = package_info["versions"]
        for ver_str, datas in versions.items():
            # Strip leading 'v' if present
            if ver_str.startswith("v"):
                ver_str = ver_str[1:]

            # Skip non-semver strings
            if not re.match(r"^[\d.]+$", ver_str):
                continue

            current_version = version.parse(ver_str)
            if current_version > latest:
                latest = current_version
                dependency.latest = ver_str
                dependency.license = datas.get("license", [])
                dependency.homepage = datas.get("homepage")
                dependency.zip = datas.get("dist", {}).get("url")

        if dependency.version == dependency.latest:
            dependency.status = "latest"
        else:
            dependency.status = "outdated"
        return dependency
